#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
import logging
import os
import sys
from argparse import Namespace
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import nibabel as nib
import numpy as np
from numpy.typing import NDArray

import rapidtide.io as tide_io
import rapidtide.util as tide_util
import rapidtide.workflows.parser_funcs as pf


# Create a sentinel.
# from https://stackoverflow.com/questions/58594956/find-out-which-arguments-were-passed-explicitly-in-argparse
class _Sentinel:
    pass


sentinel = _Sentinel()

LGR = logging.getLogger(__name__)

# Some default settings
DEFAULT_SPATIALFILT = -1
DEFAULT_HISTLEN = 101
DEFAULT_DETREND_ORDER = 3
DEFAULT_INITREGRESSOR_PCACOMPONENTS = 0.8
DEFAULT_CORRMASK_THRESHPCT = 1.0
DEFAULT_MUTUALINFO_SMOOTHINGTIME = 3.0
DEFAULT_LAGMIN = -30.0
DEFAULT_LAGMAX = 30.0
DEFAULT_SIGMAMAX = 1000.0
DEFAULT_SIGMAMIN = 0.0
DEFAULT_DESPECKLE_PASSES = 4
DEFAULT_DESPECKLE_THRESH = 5.0
DEFAULT_PASSES = 3
DEFAULT_LAGMIN_THRESH = 0.25
DEFAULT_LAGMAX_THRESH = 3.0
DEFAULT_AMPTHRESH = 0.3
DEFAULT_PICKLEFT_THRESH = 0.33
DEFAULT_SIGMATHRESH = 100.0
DEFAULT_MAXPASSES = 15
DEFAULT_REFINE_TYPE = "pca"
DEFAULT_INTERPTYPE = "univariate"
DEFAULT_WINDOW_TYPE = "hamming"
DEFAULT_INITREGRESSOR_METHOD = "sum"
DEFAULT_CORRWEIGHTING = "phat"
DEFAULT_CORRTYPE = "linear"
DEFAULT_SIMILARITYMETRIC = "correlation"
DEFAULT_PEAKFIT_TYPE = "gauss"
DEFAULT_REFINE_PRENORM = "var"
DEFAULT_REFINE_WEIGHTING = "None"
DEFAULT_REFINE_PCACOMPONENTS = 0.8
DEFAULT_REGRESSIONFILTDERIVS = 0

DEFAULT_DENOISING_LAGMIN = -10.0
DEFAULT_DENOISING_LAGMAX = 10.0
DEFAULT_DENOISING_PASSES = 3
DEFAULT_DENOISING_DESPECKLE_PASSES = 4
DEFAULT_DENOISING_PEAKFITTYPE = "gauss"
DEFAULT_DENOISING_SPATIALFILT = -1

DEFAULT_DELAYMAPPING_LAGMIN = -10.0
DEFAULT_DELAYMAPPING_LAGMAX = 30.0
DEFAULT_DELAYMAPPING_PASSES = 3
DEFAULT_DELAYMAPPING_DESPECKLE_PASSES = 4
DEFAULT_DELAYMAPPING_SPATIALFILT = 2.5

DEFAULT_CVRMAPPING_LAGMIN = -5.0
DEFAULT_CVRMAPPING_LAGMAX = 20.0
DEFAULT_CVRMAPPING_FILTER_LOWERPASS = 0.0
DEFAULT_CVRMAPPING_FILTER_UPPERPASS = 0.01
DEFAULT_CVRMAPPING_DESPECKLE_PASSES = 4

DEFAULT_OUTPUTLEVEL = "normal"

DEFAULT_SLFONOISEAMP_WINDOWSIZE = 40.0

DEFAULT_COARSEDELAY_TYPE = "simfunc"
DEFAULT_RIPTIDESTEP = 2.0

DEFAULT_PATCHTHRESH = 3.0
DEFAULT_REFINEDELAYMINDELAY = -5.0
DEFAULT_REFINEDELAYMAXDELAY = 5.0
DEFAULT_REFINEDELAYNUMPOINTS = 501
DEFAULT_DELAYOFFSETSPATIALFILT = -1

DEFAULT_PATCHMINSIZE = 10
DEFAULT_PATCHFWHM = 5

DEFAULT_PREWHITEN_LAGS = -1


def _get_parser() -> Any:
    """
    Set up the argument parser for rapidtide command-line interface.

    This function configures all available command-line arguments for the rapidtide
    tool, organizing them into logical groups for easy reference and use.

    Returns
    -------
    argparse.ArgumentParser
        Configured argument parser with all rapidtide options added

    Notes
    -----
    The parser includes several groups of options:
    - Input/Output options
    - Processing options
    - Performance options
    - Miscellaneous options
    - Experimental options (not fully tested)
    - Deprecated options (will be removed in future versions)
    - Debugging options (for development and troubleshooting)

    Examples
    --------
    >>> parser = setup_parser()
    >>> args = parser.parse_args()
    >>> print(args.version)
    """
    parser = argparse.ArgumentParser(
        prog="rapidtide",
        description=("Perform a RIPTiDe time delay analysis on a dataset."),
        allow_abbrev=False,
    )

    # Required arguments
    parser.add_argument(
        "in_file",
        type=lambda x: pf.is_valid_file(parser, x),
        help="The input data file (BOLD fMRI file or NIRS text file).",
    )
    parser.add_argument(
        "outputname",
        type=str,
        help=(
            "The root name for the output files.  "
            "For BIDS compliance, this can only contain valid BIDS entities "
            "from the source data."
        ),
    )

    # Analysis types
    analysis_type = parser.add_argument_group(
        title="Analysis type",
        description=(
            "Single arguments that set several parameter values, tailored to particular analysis types. "
            "Any parameter set by an analysis type can be overridden "
            "by setting that parameter explicitly. "
            "Analysis types are mutually exclusive with one another."
        ),
    ).add_mutually_exclusive_group()
    analysis_type.add_argument(
        "--denoising",
        dest="denoising",
        action="store_true",
        help=(
            "Preset for hemodynamic denoising - this is a macro that "
            f"sets searchrange=({DEFAULT_DENOISING_LAGMIN}, {DEFAULT_DENOISING_LAGMAX}), "
            f"passes={DEFAULT_DENOISING_PASSES}, despeckle_passes={DEFAULT_DENOISING_DESPECKLE_PASSES}, "
            f"refineoffset=True, peakfittype={DEFAULT_DENOISING_PEAKFITTYPE}, "
            f"gausssigma={DEFAULT_DENOISING_SPATIALFILT}, nofitfilt=True, dolinfitfilt=True. "
            "Any of these options can be overridden with the appropriate "
            "additional arguments."
        ),
        default=False,
    )
    analysis_type.add_argument(
        "--delaymapping",
        dest="delaymapping",
        action="store_true",
        help=(
            "Preset for delay mapping analysis - this is a macro that "
            f"sets searchrange=({DEFAULT_DELAYMAPPING_LAGMIN}, {DEFAULT_DELAYMAPPING_LAGMAX}), "
            f"passes={DEFAULT_DELAYMAPPING_PASSES}, despeckle_passes={DEFAULT_DELAYMAPPING_DESPECKLE_PASSES}, "
            f"gausssigma={DEFAULT_DELAYMAPPING_SPATIALFILT}, "
            "refineoffset=True, refinedelay=True, outputlevel='normal'. "
            "Any of these options can be overridden with the appropriate "
            "additional arguments."
        ),
        default=False,
    )
    analysis_type.add_argument(
        "--CVR",
        dest="docvrmap",
        action="store_true",
        help=(
            "Preset for calibrated CVR mapping.  Given an input regressor that represents some measured "
            "quantity over time (e.g. mmHg CO2 in the EtCO2 trace), rapidtide will calculate and output a map of percent "
            "BOLD change in units of the input regressor.  To do this, this sets: "
            f"passes=1, despeckle_passes={DEFAULT_CVRMAPPING_DESPECKLE_PASSES}, "
            f"searchrange=({DEFAULT_CVRMAPPING_LAGMIN}, {DEFAULT_CVRMAPPING_LAGMAX}), "
            f"filterfreqs=({DEFAULT_CVRMAPPING_FILTER_LOWERPASS}, {DEFAULT_CVRMAPPING_FILTER_UPPERPASS}), "
            "and calculates a voxelwise regression fit using the optimally delayed "
            "input regressor and the percent normalized, demeaned BOLD data as inputs. This map is output as "
            "(XXX_desc-CVR_map.nii.gz).  If no input regressor is supplied, this will generate an error.  "
            "These options can be overridden with the appropriate additional arguments."
        ),
        default=False,
    )
    analysis_type.add_argument(
        "--globalpreselect",
        dest="initregressorpreselect",
        action="store_true",
        help=(
            "Treat this run as an initial pass to locate good candidate voxels for global mean "
            "regressor generation.  This sets: passes=1, despecklepasses=0, "
            "refinedespeckle=False, outputlevel='normal', dolinfitfilt=False, saveintermediatemaps=False."
        ),
        default=False,
    )

    # Macros
    macros = parser.add_argument_group(
        title="Macros",
        description=(
            "Single arguments that change default values for many "
            "arguments. "
            "Macros override individually set parameters. "
        ),
    ).add_mutually_exclusive_group()
    macros.add_argument(
        "--venousrefine",
        dest="venousrefine",
        action="store_true",
        help=(
            "This is a macro that sets lagminthresh=2.5, "
            "lagmaxthresh=6.0, ampthresh=0.5, and "
            "refineupperlag to bias refinement towards "
            "voxels in the draining vasculature for an "
            "fMRI scan."
        ),
        default=False,
    )
    macros.add_argument(
        "--nirs",
        dest="nirs",
        action="store_true",
        help=(
            "This is a NIRS analysis - this is a macro that "
            "sets nothresh, dataiszeromean=True, refineprenorm=var, ampthresh=0.7, and "
            "lagminthresh=0.1. "
        ),
        default=False,
    )

    anatomy = parser.add_argument_group(
        title="Anatomic information",
        description=(
            "These options allow you to tailor the analysis with some anatomic constraints.  You don't need to supply "
            "any of them, but if you do, rapidtide will try to make intelligent processing decisions based on "
            "these maps.  Any individual masks set with anatomic information will be overridden if you specify "
            "that mask directly."
        ),
    )
    anatomy.add_argument(
        "--brainmask",
        dest="brainmaskincludespec",
        metavar="MASK[:VALSPEC]",
        help=(
            "This specifies the valid brain voxels.  No voxels outside of this mask will be used for global mean "
            "calculation, correlation, refinement, offset calculation, or denoising. "
            "If VALSPEC is given, only voxels in the mask with integral values listed in VALSPEC are used, otherwise "
            "voxels with value > 0.1 are used.  If this option is set, "
            "rapidtide will limit the include mask used to 1) calculate the initial global mean regressor, "
            "2) decide which voxels in which to calculate delays, "
            "3) refine the regressor at the end of each pass, 4) determine the zero time offset value, and 5) process "
            "to remove sLFO signal. "
            "Setting --initregressorinclude, --corrmaskinclude, --refineinclude or --offsetinclude explicitly will "
            "override this for the given include mask."
        ),
        default=None,
    )
    anatomy.add_argument(
        "--graymattermask",
        dest="graymatterincludespec",
        metavar="MASK[:VALSPEC]",
        help=(
            "This specifies a gray matter mask registered to the input functional data.  "
            "If VALSPEC is given, only voxels in the mask with integral values listed in VALSPEC are used, otherwise "
            "voxels with value > 0.1 are used.  If this option is set, "
            "rapidtide will use voxels in the gray matter mask to calculate the initial global mean regressor. "
            "Setting --initregressorinclude  explicitly will override this for "
            "the given include mask."
        ),
        default=None,
    )
    anatomy.add_argument(
        "--whitemattermask",
        dest="whitematterincludespec",
        metavar="MASK[:VALSPEC]",
        help=(
            "This specifies a white matter mask registered to the input functional data.  "
            "If VALSPEC is given, only voxels in the mask with integral values listed in VALSPEC are used, otherwise "
            "voxels with value > 0.1 are used.  "
            "This will be used to calculate the white matter timecourse before running rapidtide, and after sLFO "
            "filtering (if enabled)."
        ),
        default=None,
    )
    anatomy.add_argument(
        "--csfmask",
        dest="csfincludespec",
        metavar="MASK[:VALSPEC]",
        help=(
            "This specifies a CSF mask registered to the input functional data.  "
            "If VALSPEC is given, only voxels in the mask with integral values listed in VALSPEC are used, otherwise "
            "voxels with value > 0.1 are used.  "
            "This will be used to calculate the CSF timecourse before running rapidtide, and after sLFO "
            "filtering (if enabled)."
        ),
        default=None,
    )

    # Preprocessing options
    preproc = parser.add_argument_group("Preprocessing options")
    realtr = preproc.add_mutually_exclusive_group()
    realtr.add_argument(
        "--datatstep",
        dest="realtr",
        action="store",
        metavar="TSTEP",
        type=lambda x: pf.is_float(parser, x, minval=0.0),
        help=(
            "Set the timestep of the data file to TSTEP. "
            "This will override the TR in an "
            "fMRI file. NOTE: if using data from a text "
            "file, for example with NIRS data, using one "
            "of these options is mandatory. "
        ),
        default="auto",
    )
    realtr.add_argument(
        "--datafreq",
        dest="realtr",
        action="store",
        metavar="FREQ",
        type=lambda x: pf.invert_float(parser, x),
        help=(
            "Set the timestep of the data file to 1/FREQ. "
            "This will override the TR in an "
            "fMRI file. NOTE: if using data from a text "
            "file, for example with NIRS data, using one "
            "of these options is mandatory. "
        ),
        default="auto",
    )
    preproc.add_argument(
        "--noantialias",
        dest="antialias",
        action="store_false",
        help="Disable antialiasing filter. ",
        default=True,
    )
    preproc.add_argument(
        "--invert",
        dest="invertregressor",
        action="store_true",
        help=("Invert the sign of the regressor before processing."),
        default=False,
    )
    preproc.add_argument(
        "--interptype",
        dest="interptype",
        action="store",
        type=str,
        choices=["univariate", "cubic", "quadratic"],
        help=(
            "Use specified interpolation type. Options "
            'are "cubic", "quadratic", and "univariate". '
            f"Default is {DEFAULT_INTERPTYPE}. "
        ),
        default=DEFAULT_INTERPTYPE,
    )
    preproc.add_argument(
        "--offsettime",
        dest="offsettime",
        action="store",
        type=float,
        metavar="OFFSETTIME",
        help="Apply offset OFFSETTIME to the lag regressors.",
        default=0.0,
    )
    preproc.add_argument(
        "--autosync",
        dest="autosync",
        action="store_true",
        help=(
            "Estimate and apply the initial offsettime of an external "
            "regressor using the global crosscorrelation. "
            "Overrides offsettime if present."
        ),
        default=False,
    )
    preproc.add_argument(
        "--dataiszeromean",
        dest="dataiszeromean",
        action="store_true",
        help=(
            "Assume that the fMRI data is zero mean (this will be the case if you used AFNI for preprocessing).  "
            "This affects how masks are generated.  Rapidtide will attempt to detect this, but set explicitly "
            "if you know this is the case."
        ),
        default=False,
    )

    # Add filter options
    pf.addfilteropts(parser, filtertarget="data and regressors", details=True)

    # Add permutation options
    pf.addpermutationopts(parser)

    # add window options
    pf.addwindowopts(parser, windowtype=DEFAULT_WINDOW_TYPE)

    preproc.add_argument(
        "--detrendorder",
        dest="detrendorder",
        action="store",
        type=lambda x: pf.is_int(parser, x, minval=0),
        metavar="ORDER",
        help=(f"Set order of trend removal (0 to disable). Default is {DEFAULT_DETREND_ORDER}."),
        default=DEFAULT_DETREND_ORDER,
    )
    preproc.add_argument(
        "--spatialfilt",
        dest="gausssigma",
        action="store",
        type=float,
        metavar="GAUSSSIGMA",
        help=(
            "Spatially filter fMRI data prior to analysis "
            "using GAUSSSIGMA in mm.  Set GAUSSSIGMA negative "
            "to have rapidtide set it to half the mean voxel "
            "dimension (a rule of thumb for a good value). Set to 0 to disable."
        ),
        default=DEFAULT_SPATIALFILT,
    )
    preproc.add_argument(
        "--premask",
        dest="premask",
        action="store_true",
        help=(
            "Apply masking prior to spatial filtering to limit extracerebral sources (requires --brainmask)"
        ),
        default=False,
    )
    preproc.add_argument(
        "--premasktissueonly",
        dest="premasktissueonly",
        action="store_true",
        help=(
            "Apply more stringent masking prior to spatial filtering, removing CSF voxels (requires --graymattermask and --whitemattermask)."
        ),
        default=False,
    )
    preproc.add_argument(
        "--globalmean",
        dest="useinitregressorref",
        action="store_true",
        help=(
            "Generate a global mean regressor and use that as the reference "
            "regressor.  If no external regressor is specified, this "
            "is enabled by default."
        ),
        default=False,
    )
    preproc.add_argument(
        "--globalmeaninclude",
        dest="initregressorincludespec",
        metavar="MASK[:VALSPEC]",
        help=(
            "Only use voxels in mask file NAME for the initial regressor "
            "generation (if VALSPEC is given, only voxels "
            "with integral values listed in VALSPEC are used)."
        ),
        default=None,
    )
    preproc.add_argument(
        "--globalmeanexclude",
        dest="initregressorexcludespec",
        metavar="MASK[:VALSPEC]",
        help=(
            "Do not use voxels in mask file NAME for the initial regressor "
            "generation (if VALSPEC is given, only voxels "
            "with integral values listed in VALSPEC are excluded)."
        ),
        default=None,
    )
    preproc.add_argument(
        "--motionfile",
        dest="motionfilespec",
        metavar="MOTFILE",
        help=(
            "Read 6 columns of motion regressors out of MOTFILE file (.par or BIDS .json) "
            "(with timepoints rows) and regress them and/or their derivatives "
            "out of the data prior to analysis. "
        ),
        default=None,
    )
    preproc.add_argument(
        "--motpowers",
        dest="mot_power",
        metavar="N",
        type=lambda x: pf.is_int(parser, x, minval=1),
        help=(
            "Include powers of each motion regressor up to order N. Default is 1 (no expansion). "
        ),
        default=1,
    )
    preproc.add_argument(
        "--nomotderiv",
        dest="mot_deriv",
        action="store_false",
        help=(
            "Do not use derivatives in motion regression.  Default is to use temporal derivatives."
        ),
        default=True,
    )
    preproc.add_argument(
        "--confoundfile",
        dest="confoundfilespec",
        metavar="CONFFILE",
        help=(
            "Read additional (non-motion) confound regressors out of CONFFILE file (which can be any type of "
            "multicolumn text file "
            "rapidtide reads as long as data is sampled at TR with timepoints rows).  Optionally do power expansion "
            "and/or calculate derivatives prior to regression. "
        ),
        default=None,
    )
    preproc.add_argument(
        "--confoundpowers",
        dest="confound_power",
        metavar="N",
        type=lambda x: pf.is_int(parser, x, minval=1),
        help=(
            "Include powers of each confound regressor up to order N. Default is 1 (no expansion). "
        ),
        default=1,
    )
    preproc.add_argument(
        "--noconfoundderiv",
        dest="confound_deriv",
        action="store_false",
        help=(
            "Do not use derivatives in confound regression.  Default is to use temporal derivatives."
        ),
        default=True,
    )
    preproc.add_argument(
        "--noconfoundorthogonalize",
        dest="orthogonalize",
        action="store_false",
        help=(
            "Do not orthogonalize confound regressors prior to regressing them out of the data. "
        ),
        default=True,
    )
    preproc.add_argument(
        "--globalsignalmethod",
        dest="initregressorsignalmethod",
        action="store",
        type=str,
        choices=["sum", "meanscale", "pca", "random"],
        help=(
            "The method for constructing the initial signal regressor - straight summation, "
            "mean scaling each voxel prior to summation, MLE PCA of the voxels in the initial regressor mask, "
            "or initializing using random noise. "
            f'Default is "{DEFAULT_INITREGRESSOR_METHOD}."'
        ),
        default=DEFAULT_INITREGRESSOR_METHOD,
    )
    preproc.add_argument(
        "--globalpcacomponents",
        dest="initregressorpcacomponents",
        action="store",
        type=float,
        metavar="VALUE",
        help=(
            "Number of PCA components used for estimating the initial regressor.  If VALUE >= 1, will retain this"
            "many components.  If "
            "0.0 < VALUE < 1.0, enough components will be retained to explain the fraction VALUE of the "
            "total variance. If VALUE is negative, the number of components will be to retain will be selected "
            f"automatically using the MLE method.  Default is {DEFAULT_INITREGRESSOR_PCACOMPONENTS}."
        ),
        default=DEFAULT_INITREGRESSOR_PCACOMPONENTS,
    )
    preproc.add_argument(
        "--slicetimes",
        dest="slicetimes",
        action="store",
        type=lambda x: pf.is_valid_file(parser, x),
        metavar="FILE",
        help=("Apply offset times from FILE to each slice in the dataset."),
        default=None,
    )
    preproc.add_argument(
        "--numskip",
        dest="preprocskip",
        action="store",
        type=lambda x: pf.is_int(parser, x, minval=0),
        metavar="SKIP",
        help=(
            "SKIP TRs were previously deleted during "
            "preprocessing (e.g. if you have done your preprocessing "
            "in FSL and set dummypoints to a nonzero value.) Default is 0. "
        ),
        default=0,
    )
    preproc.add_argument(
        "--timerange",
        dest="timerange",
        action="store",
        nargs=2,
        type=int,
        metavar=("START", "END"),
        help=(
            "Limit analysis to data between timepoints "
            "START and END in the fmri file. If END is set to -1, "
            "analysis will go to the last timepoint.  Negative values "
            "of START will be set to 0. Default is to use all timepoints."
        ),
        default=(-1, -1),
    )
    preproc.add_argument(
        "--tincludemask",
        dest="tincludemaskname",
        action="store",
        type=lambda x: pf.is_valid_file(parser, x),
        metavar="FILE",
        help=(
            "Only correlate during epochs specified "
            "in MASKFILE (NB: each line of FILE "
            "contains the time and duration of an "
            "epoch to include."
        ),
        default=None,
    )
    preproc.add_argument(
        "--texcludemask",
        dest="texcludemaskname",
        action="store",
        type=lambda x: pf.is_valid_file(parser, x),
        metavar="FILE",
        help=(
            "Do not correlate during epochs specified "
            "in MASKFILE (NB: each line of FILE "
            "contains the time and duration of an "
            "epoch to exclude."
        ),
        default=None,
    )
    preproc.add_argument(
        "--nothresh",
        dest="nothresh",
        action="store_true",
        help=("Disable voxel intensity threshold (especially useful for NIRS " "data)."),
        default=False,
    )

    # Correlation options
    corr = parser.add_argument_group("Correlation options")
    corr.add_argument(
        "--oversampfac",
        dest="oversampfactor",
        action="store",
        type=int,
        metavar="OVERSAMPFAC",
        help=(
            "Oversample the fMRI data by the following "
            "integral factor.  Set to -1 for automatic selection (default)."
        ),
        default=-1,
    )
    corr.add_argument(
        "--regressor",
        dest="regressorfile",
        action="store",
        type=lambda x: pf.is_valid_file(parser, x),
        metavar="FILE",
        help=(
            "Read the initial probe regressor from file FILE (if not "
            "specified, generate and use the global regressor)."
        ),
        default=None,
    )

    reg_group = corr.add_mutually_exclusive_group()
    reg_group.add_argument(
        "--regressorfreq",
        dest="inputfreq",
        action="store",
        type=lambda x: pf.is_float(parser, x, minval=0.0),
        metavar="FREQ",
        help=(
            "Probe regressor in file has sample "
            "frequency FREQ (default is 1/tr) "
            "NB: --regressorfreq and --regressortstep) "
            "are two ways to specify the same thing."
        ),
        default="auto",
    )
    reg_group.add_argument(
        "--regressortstep",
        dest="inputfreq",
        action="store",
        type=lambda x: pf.invert_float(parser, x),
        metavar="TSTEP",
        help=(
            "Probe regressor in file has sample "
            "frequency FREQ (default is 1/tr) "
            "NB: --regressorfreq and --regressortstep) "
            "are two ways to specify the same thing."
        ),
        default="auto",
    )

    corr.add_argument(
        "--regressorstart",
        dest="inputstarttime",
        action="store",
        type=float,
        metavar="START",
        help=(
            "The time delay in seconds into the regressor "
            "file, corresponding in the first TR of the fMRI "
            "file (default is 0.0)."
        ),
        default=None,
    )
    corr.add_argument(
        "--corrweighting",
        dest="corrweighting",
        action="store",
        type=str,
        choices=["None", "phat", "liang", "eckart", "regressor"],
        help=(
            "Method to use for cross-correlation weighting. "
            "'None' performs an unweighted correlation. "
            "'phat' weights the correlation by the magnitude of the product of the timecourse's FFTs. "
            "'liang' weights the correlation by the sum of the magnitudes of the timecourse's FFTs. "
            "'eckart' weights the correlation by the product of the magnitudes of the timecourse's FFTs. "
            "'regressor' weights the correlation by the magnitude of the sLFO regressor FFT. "
            f'Default is "{DEFAULT_CORRWEIGHTING}".'
        ),
        default=DEFAULT_CORRWEIGHTING,
    )
    corr.add_argument(
        "--corrtype",
        dest="corrtype",
        action="store",
        type=str,
        choices=["linear", "circular"],
        help=("Cross-correlation type (linear or circular). " f'Default is "{DEFAULT_CORRTYPE}".'),
        default=DEFAULT_CORRTYPE,
    )

    mask_group = corr.add_mutually_exclusive_group()
    mask_group.add_argument(
        "--corrmaskthresh",
        dest="corrmaskthreshpct",
        action="store",
        type=float,
        metavar="PCT",
        help=(
            "Do correlations in voxels where the mean "
            "exceeds this percentage of the robust max. "
            f"Default is {DEFAULT_CORRMASK_THRESHPCT}. "
        ),
        default=DEFAULT_CORRMASK_THRESHPCT,
    )
    mask_group.add_argument(
        "--corrmask",
        dest="corrmaskincludespec",
        metavar="MASK[:VALSPEC]",
        help=(
            "Only do correlations in nonzero voxels in NAME "
            "(if VALSPEC is given, only voxels "
            "with integral values listed in VALSPEC are used). "
        ),
        default=None,
    )
    corr.add_argument(
        "--similaritymetric",
        dest="similaritymetric",
        action="store",
        type=str,
        choices=["correlation", "mutualinfo", "hybrid", "riptide"],
        help=(
            "Similarity metric for finding delay values.  "
            'Choices are "correlation", "mutualinfo", "hybrid", and "riptide". '
            f"Default is {DEFAULT_SIMILARITYMETRIC}."
        ),
        default=DEFAULT_SIMILARITYMETRIC,
    )
    corr.add_argument(
        "--mutualinfosmoothingtime",
        dest="smoothingtime",
        action="store",
        type=float,
        metavar="TAU",
        help=(
            "Time constant of a temporal smoothing function to apply to the "
            "mutual information function. "
            f"Default is {DEFAULT_MUTUALINFO_SMOOTHINGTIME} seconds.  "
            "TAU <=0.0 disables smoothing."
        ),
        default=DEFAULT_MUTUALINFO_SMOOTHINGTIME,
    )
    corr.add_argument(
        "--simcalcrange",
        dest="simcalcrange",
        action="store",
        nargs=2,
        type=int,
        metavar=("START", "END"),
        help=(
            "Limit correlation calculation to data between timepoints "
            "START and END in the fmri file. If END is set to -1, "
            "analysis will go to the last timepoint.  Negative values "
            "of START will be set to 0. Default is to use all timepoints. "
            "NOTE: these offsets are relative to the start of the "
            "dataset AFTER any trimming done with '--timerange'."
        ),
        default=(-1, -1),
    )

    # Correlation fitting options
    corr_fit = parser.add_argument_group("Correlation fitting options")

    corr_fit.add_argument(
        "--initialdelay",
        dest="initialdelayvalue",
        type=lambda x: pf.is_valid_file_or_float(parser, x),
        metavar="DELAY",
        help=(
            "Start the analysis with predefined delays.  If DELAY is a filename, use it as an initial delay map; "
            "if DELAY is a float, initialize all voxels to a  delay of DELAY seconds."
        ),
        default=0.0,
    )

    fixdelay = corr_fit.add_mutually_exclusive_group()
    fixdelay.add_argument(
        "--searchrange",
        dest="lag_extrema",
        action=pf.IndicateSpecifiedAction,
        nargs=2,
        type=float,
        metavar=("LAGMIN", "LAGMAX"),
        help=(
            "Limit fit to a range of lags from LAGMIN to "
            f"LAGMAX.  Default is {DEFAULT_LAGMIN} to {DEFAULT_LAGMAX} seconds. "
        ),
        default=(DEFAULT_LAGMIN, DEFAULT_LAGMAX),
    )
    fixdelay.add_argument(
        "--nodelayfit",
        dest="fixdelay",
        action="store_true",
        help=("Fix the delay in every voxel to its initial value.  Do not perform any fitting."),
        default=False,
    )
    corr_fit.add_argument(
        "--sigmalimit",
        dest="widthmax",
        action="store",
        type=float,
        metavar="SIGMALIMIT",
        help=(
            "Reject lag fits with linewidth wider than "
            f"SIGMALIMIT seconds. Default is {DEFAULT_SIGMAMAX} seconds."
        ),
        default=DEFAULT_SIGMAMAX,
    )
    corr_fit.add_argument(
        "--bipolar",
        dest="bipolar",
        action="store_true",
        help=("Bipolar mode - match peak correlation ignoring sign."),
        default=False,
    )
    corr_fit.add_argument(
        "--nofitfilt",
        dest="zerooutbadfit",
        action="store_false",
        help=("Do not zero out peak fit values if fit fails."),
        default=True,
    )
    corr_fit.add_argument(
        "--peakfittype",
        dest="peakfittype",
        action="store",
        type=str,
        choices=["gauss", "fastgauss", "quad", "fastquad", "COM", "None"],
        help=(
            "Method for fitting the peak of the similarity function "
            '"gauss" performs a Gaussian fit, and is most accurate. '
            '"quad" and "fastquad" use a quadratic fit, '
            "which is faster, but not as well tested. "
            f'Default is "{DEFAULT_PEAKFIT_TYPE}".'
        ),
        default=DEFAULT_PEAKFIT_TYPE,
    )
    corr_fit.add_argument(
        "--despecklepasses",
        dest="despeckle_passes",
        action=pf.IndicateSpecifiedAction,
        type=lambda x: pf.is_int(parser, x, minval=0),
        metavar="PASSES",
        help=(
            "Detect and refit suspect correlations to "
            "disambiguate peak locations in PASSES "
            f"passes.  Default is to perform {DEFAULT_DESPECKLE_PASSES} passes. "
            "Set to 0 to disable."
        ),
        default=DEFAULT_DESPECKLE_PASSES,
    )
    corr_fit.add_argument(
        "--despecklethresh",
        dest="despeckle_thresh",
        action="store",
        type=float,
        metavar="VAL",
        help=(
            "Refit correlation if median discontinuity "
            "magnitude exceeds VAL. "
            f"Default is {DEFAULT_DESPECKLE_THRESH} seconds."
        ),
        default=DEFAULT_DESPECKLE_THRESH,
    )

    # Regressor refinement options
    reg_ref = parser.add_argument_group("Regressor refinement options")
    reg_ref.add_argument(
        "--refineprenorm",
        dest="refineprenorm",
        action="store",
        type=str,
        choices=["None", "mean", "var", "std", "invlag"],
        help=(
            "Apply TYPE prenormalization to each "
            "timecourse prior to refinement. "
            f'Default is "{DEFAULT_REFINE_PRENORM}".'
        ),
        default=DEFAULT_REFINE_PRENORM,
    )
    reg_ref.add_argument(
        "--refineweighting",
        dest="refineweighting",
        action="store",
        type=str,
        choices=["None", "NIRS", "R", "R2"],
        help=(
            "Apply TYPE weighting to each timecourse prior "
            f'to refinement. Default is "{DEFAULT_REFINE_WEIGHTING}".'
        ),
        default=DEFAULT_REFINE_WEIGHTING,
    )
    reg_ref.add_argument(
        "--passes",
        dest="passes",
        action=pf.IndicateSpecifiedAction,
        type=lambda x: pf.is_int(parser, x, minval=1),
        metavar="PASSES",
        help=("Set the number of processing passes to PASSES.  " f"Default is {DEFAULT_PASSES}."),
        default=DEFAULT_PASSES,
    )
    reg_ref.add_argument(
        "--refineinclude",
        dest="refineincludespec",
        metavar="MASK[:VALSPEC]",
        help=(
            "Only use voxels in file MASK for regressor refinement "
            "(if VALSPEC is given, only voxels "
            "with integral values listed in VALSPEC are used). "
        ),
        default=None,
    )
    reg_ref.add_argument(
        "--refineexclude",
        dest="refineexcludespec",
        metavar="MASK[:VALSPEC]",
        help=(
            "Do not use voxels in file MASK for regressor refinement "
            "(if VALSPEC is given, voxels "
            "with integral values listed in VALSPEC are excluded). "
        ),
        default=None,
    )
    reg_ref.add_argument(
        "--norefinedespeckled",
        dest="refinedespeckled",
        action="store_false",
        help=("Do not use despeckled pixels in calculating the refined regressor."),
        default=True,
    )
    reg_ref.add_argument(
        "--lagminthresh",
        dest="lagminthresh",
        action="store",
        metavar="MIN",
        type=float,
        help=(
            "For refinement, exclude voxels with delays "
            f"less than MIN. Default is {DEFAULT_LAGMIN_THRESH} seconds. "
        ),
        default=DEFAULT_LAGMIN_THRESH,
    )
    reg_ref.add_argument(
        "--lagmaxthresh",
        dest="lagmaxthresh",
        action="store",
        metavar="MAX",
        type=float,
        help=(
            "For refinement, exclude voxels with delays "
            f"greater than MAX. Default is {DEFAULT_LAGMAX_THRESH} seconds. "
        ),
        default=DEFAULT_LAGMAX_THRESH,
    )
    reg_ref.add_argument(
        "--ampthresh",
        dest="ampthresh",
        action="store",
        metavar="AMP",
        type=float,
        help=(
            "For refinement, exclude voxels with correlation "
            f"coefficients less than AMP (default is {DEFAULT_AMPTHRESH}).  "
            "NOTE: ampthresh will automatically be set to the p<0.05 "
            "significance level determined by the --numnull option if NREPS "
            "is set greater than 0 and this is not manually specified."
        ),
        default=-1.0,
    )
    reg_ref.add_argument(
        "--sigmathresh",
        dest="sigmathresh",
        action="store",
        metavar="SIGMA",
        type=float,
        help=(
            "For refinement, exclude voxels with widths "
            f"greater than SIGMA seconds. Default is {DEFAULT_SIGMATHRESH} seconds."
        ),
        default=DEFAULT_SIGMATHRESH,
    )
    reg_ref.add_argument(
        "--offsetinclude",
        dest="offsetincludespec",
        metavar="MASK[:VALSPEC]",
        help=(
            "Only use voxels in file MASK for determining the zero time offset value "
            "(if VALSPEC is given, only voxels "
            "with integral values listed in VALSPEC are used). "
        ),
        default=None,
    )
    reg_ref.add_argument(
        "--offsetexclude",
        dest="offsetexcludespec",
        metavar="MASK[:VALSPEC]",
        help=(
            "Do not use voxels in file MASK for determining the zero time offset value "
            "(if VALSPEC is given, voxels "
            "with integral values listed in VALSPEC are excluded). "
        ),
        default=None,
    )
    reg_ref.add_argument(
        "--norefineoffset",
        dest="refineoffset",
        action="store_false",
        help=("Disable realigning refined regressor to zero lag."),
        default=True,
    )
    reg_ref.add_argument(
        "--nopickleft",
        dest="pickleft",
        action="store_false",
        help=("Disables selecting the leftmost delay peak when setting the refine offset."),
        default=True,
    )
    reg_ref.add_argument(
        "--pickleftthresh",
        dest="pickleftthresh",
        action="store",
        metavar="THRESH",
        type=float,
        help=(
            "Threshold value (fraction of maximum) in a histogram "
            f"to be considered the start of a peak.  Default is {DEFAULT_PICKLEFT_THRESH}."
        ),
        default=DEFAULT_PICKLEFT_THRESH,
    )
    reg_ref.add_argument(
        "--sLFOnoiseampwindow",
        dest="sLFOnoiseampwindow",
        action="store",
        metavar="SECONDS",
        type=float,
        help=(
            "Width of the averaging window for filtering the RMS sLFO waveform to calculate "
            f"amplitude change over time.  Default is {DEFAULT_SLFONOISEAMP_WINDOWSIZE}."
        ),
        default=DEFAULT_SLFONOISEAMP_WINDOWSIZE,
    )

    refine = reg_ref.add_mutually_exclusive_group()
    refine.add_argument(
        "--refineupperlag",
        dest="lagmaskside",
        action="store_const",
        const="upper",
        help=("Only use positive lags for regressor refinement."),
        default="both",
    )
    refine.add_argument(
        "--refinelowerlag",
        dest="lagmaskside",
        action="store_const",
        const="lower",
        help=("Only use negative lags for regressor refinement."),
        default="both",
    )
    reg_ref.add_argument(
        "--refinetype",
        dest="refinetype",
        action="store",
        type=str,
        choices=["pca", "ica", "weighted_average", "unweighted_average"],
        help=(
            "Method with which to derive refined regressor. "
            f'Default is "{DEFAULT_REFINE_TYPE}".'
        ),
        default=DEFAULT_REFINE_TYPE,
    )
    reg_ref.add_argument(
        "--pcacomponents",
        dest="pcacomponents",
        action="store",
        type=float,
        metavar="VALUE",
        help=(
            "Number of PCA components used for refinement.  If VALUE >= 1, will retain this many components.  If "
            "0.0 < VALUE < 1.0, enough components will be retained to explain the fraction VALUE of the "
            "total variance. If VALUE is negative, the number of components will be to retain will be selected "
            f"automatically using the MLE method.  Default is {DEFAULT_REFINE_PCACOMPONENTS}."
        ),
        default=DEFAULT_REFINE_PCACOMPONENTS,
    )
    reg_ref.add_argument(
        "--convergencethresh",
        dest="convergencethresh",
        action="store",
        type=float,
        metavar="THRESH",
        help=(
            "Continue refinement until the MSE between regressors becomes <= THRESH.  "
            "By default, this is not set, so refinement will run for the specified number of passes. "
        ),
        default=None,
    )
    reg_ref.add_argument(
        "--maxpasses",
        dest="maxpasses",
        action="store",
        type=lambda x: pf.is_int(parser, x, minval=1),
        metavar="MAXPASSES",
        help=(
            "Terminate refinement after MAXPASSES passes, whether or not convergence has occurred. "
            f"Default is {DEFAULT_MAXPASSES}."
        ),
        default=DEFAULT_MAXPASSES,
    )

    # sLFO noise removal options
    slfofilt = parser.add_argument_group("sLFO noise removal options")
    slfofilt.add_argument(
        "--nodenoise",
        dest="dolinfitfilt",
        action="store_false",
        help=(
            "Turn off regression filtering to remove delayed "
            "regressor from each voxel (disables output of "
            "fitNorm)."
        ),
        default=True,
    )
    slfofilt.add_argument(
        "--denoisesourcefile",
        dest="denoisesourcefile",
        action="store",
        type=lambda x: pf.is_valid_file(parser, x),
        metavar="FILE",
        help=(
            "Regress delayed regressors out of FILE instead "
            "of the initial fmri file used to estimate "
            "delays."
        ),
        default=None,
    )
    slfofilt.add_argument(
        "--preservefiltering",
        dest="preservefiltering",
        action="store_true",
        help="Don't reread data prior to performing sLFO filtering.",
        default=False,
    )
    slfofilt.add_argument(
        "--regressderivs",
        dest="regressderivs",
        action="store",
        type=lambda x: pf.is_int(parser, x, minval=0),
        metavar="NDERIVS",
        help=(
            f"When doing final sLFO filtering, include derivatives up to NDERIVS order. Default is {DEFAULT_REGRESSIONFILTDERIVS}"
        ),
        default=DEFAULT_REGRESSIONFILTDERIVS,
    )
    slfofilt.add_argument(
        "--norefinedelay",
        dest="refinedelay",
        action="store_false",
        help=("Do not calculate a refined delay map using sLFO regression information."),
        default=True,
    )
    slfofilt.add_argument(
        "--nofilterwithrefineddelay",
        dest="filterwithrefineddelay",
        action="store_false",
        help=("Do not use the refined delay in sLFO filter."),
        default=True,
    )
    slfofilt.add_argument(
        "--delaypatchthresh",
        dest="delaypatchthresh",
        action="store",
        type=float,
        metavar="NUMMADs",
        help=(
            "Maximum number of robust standard deviations to permit in the offset delay refine map. "
            f"Default is {DEFAULT_PATCHTHRESH}"
        ),
        default=DEFAULT_PATCHTHRESH,
    )
    slfofilt.add_argument(
        "--delayoffsetspatialfilt",
        dest="delayoffsetgausssigma",
        action="store",
        type=float,
        metavar="GAUSSSIGMA",
        help=(
            "Spatially filter fMRI data prior to calculating delay offsets "
            "using GAUSSSIGMA in mm.  Set GAUSSSIGMA negative "
            "to have rapidtide set it to half the mean voxel "
            "dimension (a rule of thumb for a good value).  Set to 0 to disable."
        ),
        default=DEFAULT_DELAYOFFSETSPATIALFILT,
    )

    # Output options
    output = parser.add_argument_group("Output options")
    output.add_argument(
        "--outputlevel",
        dest="outputlevel",
        action="store",
        type=str,
        choices=["min", "less", "normal", "more", "max"],
        help=(
            "The level of file output produced.  'min' produces only absolutely essential files, 'less' adds in "
            "the sLFO filtered data (rather than just filter efficacy metrics), 'normal' saves what you "
            "would typically want around for interactive data exploration, "
            "'more' adds files that are sometimes useful, and 'max' outputs anything you might possibly want. "
            "Selecting 'max' will produce ~3x your input datafile size as output.  "
            f'Default is "{DEFAULT_OUTPUTLEVEL}."'
        ),
        default=DEFAULT_OUTPUTLEVEL,
    )
    output.add_argument(
        "--savelags",
        dest="savecorrtimes",
        action="store_true",
        help="Save a table of lagtimes used.",
        default=False,
    )
    output.add_argument(
        "--histlen",  # was -h
        dest="histlen",
        action="store",
        type=lambda x: pf.is_int(parser, x, minval=5),
        metavar="HISTLEN",
        help=(f"Change the histogram length to HISTLEN.  Default is {DEFAULT_HISTLEN}."),
        default=DEFAULT_HISTLEN,
    )
    output.add_argument(
        "--saveintermediatemaps",
        dest="saveintermediatemaps",
        action="store_true",
        help="Save lag times, strengths, widths, mask (and shiftedtcs, if they'd normally be saved) for each pass.",
        default=False,
    )
    output.add_argument(
        "--calccoherence",
        dest="calccoherence",
        action="store_true",
        help=("Calculate and save the coherence between the final regressor and the data."),
        default=False,
    )

    # Add version options
    pf.addversionopts(parser)

    # Performance options
    perf = parser.add_argument_group("Performance options")
    perf.add_argument(
        "--nprocs",
        dest="nprocs",
        action="store",
        type=int,
        metavar="NPROCS",
        help=(
            "Use NPROCS worker processes for multiprocessing. "
            "Setting NPROCS to less than 1 sets the number of "
            "worker processes to n_cpus (unless --reservecpu is used)."
        ),
        default=1,
    )
    perf.add_argument(
        "--reservecpu",
        dest="reservecpu",
        action="store_true",
        help=(
            "When automatically setting nprocs, reserve one CPU for "
            "process management rather than using them all for worker threads."
        ),
        default=False,
    )
    perf.add_argument(
        "--mklthreads",
        dest="mklthreads",
        action="store",
        type=lambda x: pf.is_int(parser, x, minval=1),
        metavar="MKLTHREADS",
        help=(
            "If mkl library is installed, use no more than MKLTHREADS worker "
            "threads in accelerated numpy calls.  Set to -1 to use the maximum available.  Default is 1."
        ),
        default=1,
    )
    perf.add_argument(
        "--nonumba",
        dest="nonumba",
        action="store_true",
        help=(
            "By default, numba is used if present.  Use this option to disable jit "
            "compilation with numba even if it is installed."
        ),
        default=False,
    )

    # Miscellaneous options
    misc = parser.add_argument_group("Miscellaneous options")
    misc.add_argument(
        "--noprogressbar",
        dest="showprogressbar",
        action="store_false",
        help=("Will disable showing progress bars (helpful if stdout is going to a file)."),
        default=True,
    )
    misc.add_argument(
        "--makepseudofile",
        dest="makepseudofile",
        action="store_true",
        help=("Make a simulated input file from the mean and the movingsignal."),
        default=False,
    )
    misc.add_argument(
        "--checkpoint",
        dest="checkpoint",
        action="store_true",
        help="Enable run checkpoints.",
        default=False,
    )
    misc.add_argument(
        "--spcalculation",
        dest="internalprecision",
        action="store_const",
        const="single",
        help=(
            "Use single precision for internal calculations "
            "(may be useful when RAM is limited)."
        ),
        default="double",
    )
    misc.add_argument(
        "--dpoutput",
        dest="outputprecision",
        action="store_const",
        const="double",
        help=("Use double precision for output files."),
        default="single",
    )
    misc.add_argument(
        "--spatialtolerance",
        dest="spatialtolerance",
        action="store",
        type=float,
        metavar="EPSILON",
        help=(
            "When checking to see if the spatial dimensions of two NIFTI files match, allow a relative difference "
            "of EPSILON in any dimension.  By default, this is set to 0.0, requiring an exact match. "
        ),
        default=0.0,
    )
    misc.add_argument(
        "--cifti",
        dest="isgrayordinate",
        action="store_true",
        help="Data file is a converted CIFTI.",
        default=False,
    )
    misc.add_argument(
        "--simulate",
        dest="fakerun",
        action="store_true",
        help="Simulate a run - just report command line options.",
        default=False,
    )
    misc.add_argument(
        "--displayplots",
        dest="displayplots",
        action="store_true",
        help="Display plots of interesting timecourses.",
        default=False,
    )
    misc.add_argument(
        "--nosharedmem",
        dest="sharedmem",
        action="store_false",
        help=("Disable use of shared memory for large array storage."),
        default=True,
    )
    pf.addtagopts(misc)

    # Experimental options (not fully tested, may not work)
    experimental = parser.add_argument_group(
        "Experimental options (not fully tested, or not tested at all, may not work).  Beware!"
    )
    experimental.add_argument(
        "--riptidestep",  # was -h
        dest="riptidestep",
        action="store",
        type=lambda x: pf.is_float(parser, x, maxval=5.0),
        metavar="STEP",
        help=(
            f"Timestep between RIPTiDe regressors, in seconds.  Default is {DEFAULT_RIPTIDESTEP}."
        ),
        default=DEFAULT_RIPTIDESTEP,
    )
    experimental.add_argument(
        "--refinedelayeachpass",
        dest="refinedelayeachpass",
        action="store_true",
        help=("Do delay refinement in each pass."),
        default=False,
    )
    experimental.add_argument(
        "--dofinalrefine",
        dest="dofinalrefine",
        action="store_true",
        help=("Do regressor refinement on the final pass."),
        default=False,
    )
    experimental.add_argument(
        "--sLFOfiltmask",
        dest="sLFOfiltmask",
        action="store_true",
        help=("Limit sLFO filter to fit voxels."),
        default=False,
    )
    experimental.add_argument(
        "--territorymap",
        dest="territorymap",
        metavar="MAP[:VALSPEC]",
        help=(
            "This specifies a territory map.  Each territory is a set of voxels with the same integral value.  "
            "If VALSPEC is given, only territories in the mask with integral values listed in VALSPEC are used, otherwise "
            "all nonzero voxels are used.  If this option is set, certain output measures will be summarized over "
            "each territory in the map, in addition to over the whole brain.  Some interesting territory maps might be: "
            "a gray/white/csf segmentation image, an arterial territory map, lesion area vs. healthy "
            "tissue segmentation, etc.  NB: at the moment this is just a placeholder - it doesn't do anything."
        ),
        default=None,
    )
    experimental.add_argument(
        "--wiener",
        dest="dodeconv",
        action="store_true",
        help=("Do Wiener deconvolution to find voxel transfer function."),
        default=False,
    )
    experimental.add_argument(
        "--echocancel",
        dest="echocancel",
        action="store_true",
        help=("Attempt to perform echo cancellation on current moving regressor."),
        default=False,
    )
    experimental.add_argument(
        "--autorespdelete",
        dest="respdelete",
        action="store_true",
        help=("Attempt to detect and remove respiratory signal that strays into " "the LFO band."),
        default=False,
    )
    experimental.add_argument(
        "--acfix",
        dest="fix_autocorrelation",
        action="store_true",
        help=(
            "Check probe regressor for autocorrelations in order to disambiguate peak location."
        ),
        default=False,
    )
    experimental.add_argument(
        "--negativegradient",
        dest="negativegradient",
        action="store_true",
        help=(
            "Calculate the negative gradient of the fmri data after spectral filtering "
            "so you can look for CSF flow à la https://www.biorxiv.org/content/10.1101/2021.03.29.437406v1.full. "
        ),
        default=False,
    )
    experimental.add_argument(
        "--negativegradregressor",
        dest="negativegradregressor",
        action="store_true",
        help=argparse.SUPPRESS,
        default=False,
    )
    experimental.add_argument(
        "--cleanrefined",
        dest="cleanrefined",
        action="store_true",
        help=(
            "Perform additional processing on refined "
            "regressor to remove spurious "
            "components."
        ),
        default=False,
    )
    experimental.add_argument(
        "--donotfilterinputdata",
        dest="filterinputdata",
        action="store_false",
        help=(
            "Do not filter input data prior to similarity calculation.  May make processing faster."
        ),
        default=True,
    )
    experimental.add_argument(
        "--dispersioncalc",
        dest="dodispersioncalc",
        action="store_true",
        help=("Generate extra data during refinement to allow calculation of dispersion."),
        default=False,
    )
    experimental.add_argument(
        "--patchshift",
        dest="patchshift",
        action="store_true",
        help=("Perform patch shift correction."),
        default=False,
    )
    experimental.add_argument(
        "--prewhitenregressor",
        dest="prewhitenregressor",
        action="store_true",
        help=("Prewhiten probe regressor prior to calculating correlations."),
        default=False,
    )
    experimental.add_argument(
        "--prewhitenlags",
        dest="prewhitenlags",
        action="store",
        type=lambda x: pf.is_int(parser, x, minval=0),
        metavar="LAGS",
        help=(
            f"Set number of TRs to use in prewhitening.  Set to -1 to calculate automatically. Default is {DEFAULT_PREWHITEN_LAGS}."
        ),
        default=DEFAULT_PREWHITEN_LAGS,
    )

    # Deprecated options
    deprecated = parser.add_argument_group(
        "Deprecated options.  These options will go away in the indeterminate future.  Don't use them."
    )
    deprecated.add_argument(
        "--refinedelay",
        dest="dummy",
        action="store_true",
        help=(
            "Calculate a refined delay map using regression coefficient information. ***DEPRECATED***: refinedelay is now on by default."
        ),
        default=True,
    )
    deprecated.add_argument(
        "--nolimitoutput",
        dest="limitoutput",
        action="store_false",
        help=(
            "Save some of the large and rarely used files.  "
            "***DEPRECATED***: Use '--outputlevel max' instead."
        ),
        default=True,
    )
    deprecated.add_argument(
        "--pickleft",
        dest="dummy",
        action="store_true",
        help=(
            "***DEPRECATED***: pickleft is now on by default. Use 'nopickleft' to disable it instead."
        ),
        default=True,
    )
    deprecated.add_argument(
        "--noglm",
        dest="dolinfitfilt",
        action="store_false",
        help=(
            "Turn off regression filtering to remove delayed "
            "regressor from each voxel (disables output of "
            "fitNorm)."
            "***DEPRECATED***: Use '--nodenoise' instead."
        ),
        default=True,
    )

    # Debugging options
    debugging = parser.add_argument_group(
        "Debugging options.  You probably don't want to use any of these unless I ask you to to help diagnose a problem"
    )
    debugging.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help=("Enable additional debugging output."),
        default=False,
    )
    debugging.add_argument(
        "--focaldebug",
        dest="focaldebug",
        action="store_true",
        help=("Enable targeted additional debugging output (used during development)."),
        default=False,
    )
    debugging.add_argument(
        "--verbose",
        dest="verbose",
        action="store_true",
        help=("Enable additional runtime information output."),
        default=False,
    )
    debugging.add_argument(
        "--disablecontainermemfix",
        dest="containermemfix",
        action="store_false",
        help=("Disable container memory limit setting."),
        default=True,
    )
    debugging.add_argument(
        "--alwaysmultiproc",
        dest="alwaysmultiproc",
        action="store_true",
        help=("Use the multiprocessing code path even when nprocs=1."),
        default=False,
    )
    debugging.add_argument(
        "--singleproc_confoundregress",
        dest="singleproc_confoundregress",
        action="store_true",
        help=("Force single proc path for confound regression."),
        default=False,
    )
    debugging.add_argument(
        "--singleproc_getNullDist",
        dest="singleproc_getNullDist",
        action="store_true",
        help=("Force single proc path for getNullDist."),
        default=False,
    )
    debugging.add_argument(
        "--singleproc_calcsimilarity",
        dest="singleproc_calcsimilarity",
        action="store_true",
        help=("Force single proc path for calcsimilarity."),
        default=False,
    )
    debugging.add_argument(
        "--singleproc_peakeval",
        dest="singleproc_peakeval",
        action="store_true",
        help=("Force single proc path for peakeval."),
        default=False,
    )
    debugging.add_argument(
        "--singleproc_fitcorr",
        dest="singleproc_fitcorr",
        action="store_true",
        help=("Force single proc path for fitcorr."),
        default=False,
    )
    debugging.add_argument(
        "--singleproc_refine",
        dest="singleproc_refine",
        action="store_true",
        help=("Force single proc path for refine."),
        default=False,
    )
    debugging.add_argument(
        "--singleproc_makelaggedtcs",
        dest="singleproc_makelaggedtcs",
        action="store_true",
        help=("Force single proc path for makelaggedtcs."),
        default=False,
    )
    debugging.add_argument(
        "--singleproc_regressionfilt",
        dest="singleproc_regressionfilt",
        action="store_true",
        help=("Force single proc path for regression filtering."),
        default=False,
    )
    debugging.add_argument(
        "--savecorrout",
        dest="savecorrout",
        action="store_true",
        help=("Save the corrout file even if you normally wouldn't."),
        default=False,
    )
    debugging.add_argument(
        "--isatest",
        dest="isatest",
        action="store_true",
        help=("This run of rapidtide is in a unit test."),
        default=False,
    )

    return parser


def process_args(inputargs: Optional[Any] = None) -> Tuple[Any, object]:
    """
    Compile arguments for rapidtide workflow.

    This function processes command-line arguments and sets up the configuration
    for the rapidtide workflow. It handles argument parsing, logging setup,
    command-line saving, and various parameter defaults and overrides based on
    analysis modes and macros.

    Parameters
    ----------
    inputargs : optional
        Input arguments to be processed. If None, arguments are parsed from
        sys.argv. Default is None.

    Returns
    -------
    tuple
        A tuple containing:
        - args : dict
            Dictionary of processed arguments.
        - theprefilter : object
            Preprocessing filter object.

    Notes
    -----
    The function performs the following key operations:
    1. Parses command-line arguments using `_get_parser`
    2. Sets up logging based on debug flag
    3. Saves raw and formatted command lines to files
    4. Applies default values and overrides for various parameters
    5. Handles analysis modes (delaymapping, denoising, cvrmap, etc.)
    6. Processes macros (venousrefine, nirs)
    7. Configures output options based on `outputlevel`
    8. Sets up pass options and dispersion calculation parameters
    9. Handles mask specifications and file processing
    10. Initializes filter options

    Examples
    --------
    >>> args, prefilter = process_args()
    >>> print(args['passes'])
    1
    >>> print(args['outputlevel'])
    'normal'
    """
    inargs, argstowrite = pf.setargs(_get_parser, inputargs=inputargs)
    args = vars(inargs)

    sh = logging.StreamHandler()
    if args["debug"]:
        logging.basicConfig(level=logging.DEBUG, handlers=[sh])
    else:
        logging.basicConfig(level=logging.INFO, handlers=[sh])

    # save the raw and formatted command lines
    args["commandlineargs"] = argstowrite[1:]
    thecommandline = " ".join(argstowrite)
    tide_io.writevec(np.array([thecommandline]), args["outputname"] + "_commandline.txt")
    formattedcommandline = []
    for thetoken in argstowrite[0:3]:
        formattedcommandline.append(thetoken)
    for thetoken in argstowrite[3:]:
        if thetoken[0:2] == "--":
            formattedcommandline.append(thetoken)
        else:
            formattedcommandline[-1] += " " + thetoken
    for i in range(len(formattedcommandline)):
        if i > 0:
            prefix = "    "
        else:
            prefix = ""
        if i < len(formattedcommandline) - 1:
            suffix = " \\"
        else:
            suffix = ""
        formattedcommandline[i] = prefix + formattedcommandline[i] + suffix
    tide_io.writevec(
        np.array(formattedcommandline), args["outputname"] + "_formattedcommandline.txt"
    )

    LGR.debug("\nbefore postprocessing:\n{}".format(args))

    # some tunable parameters for internal debugging
    args["dodemean"] = True
    # what fraction of the correlation window to avoid on either end when
    # fitting
    args["edgebufferfrac"] = 0.0
    # only do fits in voxels that exceed threshold
    args["enforcethresh"] = True
    # if set to the location of the first autocorrelation sidelobe,
    # this will fold back sidelobes
    args["lagmod"] = 1000.0
    # zero out peaks with correlations lower than this value
    args["lthreshval"] = 0.0
    # zero out peaks with correlations higher than this value
    args["uthreshval"] = 1.0
    # width of the reference autocorrelation function
    args["absmaxsigma"] = 10000.0
    # width of the reference autocorrelation function
    args["absminsigma"] = 0.05
    # number of MADs away from the median to consider an outlier
    args["sigdistoutlierfac"] = 10.0

    # correlation fitting
    # Peak value must be within specified range.
    # If false, allow max outside if maximum
    # correlation value is that one end of the range.
    args["hardlimit"] = True
    # The fraction of the main peak over which points are included in the peak
    args["searchfrac"] = 0.5
    args["mp_chunksize"] = 500
    args["patchminsize"] = DEFAULT_PATCHMINSIZE
    args["patchfwhm"] = DEFAULT_PATCHFWHM

    # significance estimation
    args["sighistlen"] = 1000
    if args["corrtype"] == "linear":
        args["corrpadding"] = -1
        # pf.setifnotset(args, "windowfunc", "None")
    else:
        args["corrpadding"] = 0

    # refinement options
    args["filterbeforePCA"] = True
    args["dispersioncalc_step"] = 0.50

    # autocorrelation processing
    args["check_autocorrelation"] = True
    args["acwidth"] = 0.0  # width of the reference autocorrelation function

    # delay refinement
    args["mindelay"] = DEFAULT_REFINEDELAYMINDELAY
    args["maxdelay"] = DEFAULT_REFINEDELAYMAXDELAY
    args["numpoints"] = DEFAULT_REFINEDELAYNUMPOINTS
    # diagnostic information about version
    (
        args["release_version"],
        args["git_sha"],
        args["git_date"],
        args["git_isdirty"],
    ) = tide_util.version()
    args["python_version"] = str(sys.version_info)

    # split infotags
    theobj = pf.postprocesstagopts(Namespace(**args))
    args = vars(theobj)

    # Additional argument parsing not handled by argparse
    args["passes"] = np.max([args["passes"], 1])

    args["despeckle_passes"] = np.max([args["despeckle_passes"], 0])

    if "lag_extrema_nondefault" in args.keys():
        args["lagmin_nondefault"] = True
        args["lagmax_nondefault"] = True

    args["lagmin"] = args["lag_extrema"][0]
    args["lagmax"] = args["lag_extrema"][1]

    if args["prewhitenlags"] == -1:
        args["prewhitenlags"] = int(np.max([np.fabs(args["lagmin"]), np.fabs(args["lagmax"])]))

    # set startpoint and endpoint
    args["startpoint"], args["endpoint"] = pf.parserange(args["timerange"], descriptor="timerange")

    # set simcalc startpoint and endpoint
    args["simcalcstartpoint"], args["simcalcendpoint"] = pf.parserange(
        args["simcalcrange"], descriptor="simcalcrange"
    )

    args["offsettime_total"] = args["offsettime"] + 0.0

    reg_ref_used = (
        (args["lagminthresh"] != 0.5)
        or (args["lagmaxthresh"] != 5.0)
        or (args["ampthresh"] != DEFAULT_AMPTHRESH)
        or (args["sigmathresh"] != 100.0)
        or (args["refineoffset"])
    )
    if reg_ref_used and args["passes"] == 1:
        LGR.warning(
            "One or more arguments have been set that are only "
            "relevant if performing refinement.  "
            "If you want to do refinement, set passes > 1."
        )

    if args["numestreps"] == 0:
        args["ampthreshfromsig"] = False
    else:
        args["ampthreshfromsig"] = True

    if args["ampthresh"] < 0.0:
        args["ampthresh"] = DEFAULT_AMPTHRESH
    else:
        args["ampthreshfromsig"] = False

    if args["despeckle_thresh"] != 5.0 and args["despeckle_passes"] == 0:
        args["despeckle_passes"] = 1

    if args["zerooutbadfit"]:
        args["nohistzero"] = False
    else:
        args["nohistzero"] = True

    # sort out initial delay options
    if args["fixdelay"]:
        try:
            tempinitialdelayvalue = float(args["initialdelayvalue"])
        except ValueError:
            tempinitialdelayvalue = 0.0
        args["lag_extrema"] = (
            tempinitialdelayvalue - 10.0,
            tempinitialdelayvalue + 10.0,
        )
    else:
        args["initialdelayvalue"] = None

    if args["in_file"].endswith("txt") and args["realtr"] == "auto":
        raise ValueError(
            "Either --datatstep or --datafreq must be provided if data file is a text file."
        )

    if args["realtr"] != "auto":
        fmri_tr = float(args["realtr"])
    else:
        if tide_io.checkifcifti(args["in_file"]):
            fmri_tr, dummy = tide_io.getciftitr(nib.load(args["in_file"]).header)
        else:
            fmri_tr, dummy = tide_io.fmritimeinfo(args["in_file"])
    args["realtr"] = fmri_tr

    if args["inputfreq"] == "auto":
        args["inputfreq"] = 1.0 / fmri_tr
        args["inputfreq_nondefault"] = False
    else:
        args["inputfreq_nondefault"] = True

    # initial mask values from anatomical images (before any overrides)
    for key in [
        "corrmaskincludename",
        "corrmaskincludevals",
        "initregressorincludename",
        "initregressorincludevals",
        "initregressorexcludename",
        "initregressorexcludevals",
        "refineincludename",
        "refineincludevals",
        "refineexcludename",
        "refineexcludevals",
        "offsetincludename",
        "offsetincludevals",
        "offsetexcludename",
        "offsetexcludevals",
    ]:
        args[key] = None

    # if brainmaskincludespec is set, set initregressorinclude, corrmaskinclude, refineinclude, and offsetinclude to it.
    brainmasks = ["initregressor", "corrmask", "refine", "offset"]
    if args["brainmaskincludespec"] is not None:
        (
            args["brainmaskincludename"],
            args["brainmaskincludevals"],
        ) = tide_io.processnamespec(
            args["brainmaskincludespec"], "Limiting all processing to voxels ", "in brain mask."
        )
        if not os.path.isfile(args["brainmaskincludename"]):
            raise FileNotFoundError(f"file {args['brainmaskincludename']} does not exist.")
        for masktype in brainmasks:
            (
                args[f"{masktype}includename"],
                args[f"{masktype}includevals"],
            ) = (
                args["brainmaskincludename"],
                args["brainmaskincludevals"],
            )
            print(f"setting {masktype}include mask to gray matter mask")
    else:
        args["brainmaskincludename"] = None
        args["brainmaskincludevals"] = None

    # if graymatterincludespec is set, set initregressorinclude, offsetinclude to it.
    # graymasks = ["initregressor", "offset"]
    graymasks = ["initregressor"]
    if args["graymatterincludespec"] is not None:
        (
            args["graymatterincludename"],
            args["graymatterincludevals"],
        ) = tide_io.processnamespec(
            args["graymatterincludespec"], "Including voxels where ", "in gray matter mask."
        )
        if not os.path.isfile(args["graymatterincludename"]):
            raise FileNotFoundError(f"file {args['graymatterincludename']} does not exist.")
        for masktype in graymasks:
            (
                args[f"{masktype}includename"],
                args[f"{masktype}includevals"],
            ) = (
                args["graymatterincludename"],
                args["graymatterincludevals"],
            )
            print(f"setting {masktype}include mask to gray matter mask")
    else:
        args["graymatterincludename"] = None
        args["graymatterincludevals"] = None

    if args["whitematterincludespec"] is not None:
        (
            args["whitematterincludename"],
            args["whitematterincludevals"],
        ) = tide_io.processnamespec(
            args["whitematterincludespec"], "Including voxels where ", "in white matter mask."
        )
    else:
        args["whitematterincludename"] = None
        args["whitematterincludevals"] = None

    if args["csfincludespec"] is not None:
        (
            args["csfincludename"],
            args["csfincludevals"],
        ) = tide_io.processnamespec(
            args["csfincludespec"], "Including voxels where ", "in CSF mask."
        )
    else:
        args["csfincludename"] = None
        args["csfincludevals"] = None

    # individual mask processing (including overrides)
    if args["corrmaskincludespec"] is not None:
        (
            args["corrmaskincludename"],
            args["corrmaskincludevals"],
        ) = tide_io.processnamespec(
            args["corrmaskincludespec"],
            "Including voxels where ",
            "in correlation calculations.",
        )

    if args["initregressorincludespec"] is not None:
        (
            args["initregressorincludename"],
            args["initregressorincludevals"],
        ) = tide_io.processnamespec(
            args["initregressorincludespec"],
            "Including voxels where ",
            "in initial regressor calculation.",
        )

    if args["initregressorexcludespec"] is not None:
        (
            args["initregressorexcludename"],
            args["initregressorexcludevals"],
        ) = tide_io.processnamespec(
            args["initregressorexcludespec"],
            "Excluding voxels where ",
            "from initial regressor calculation.",
        )

    if args["refineincludespec"] is not None:
        (
            args["refineincludename"],
            args["refineincludevals"],
        ) = tide_io.processnamespec(
            args["refineincludespec"], "Including voxels where ", "in refinement."
        )

    if args["refineexcludespec"] is not None:
        (
            args["refineexcludename"],
            args["refineexcludevals"],
        ) = tide_io.processnamespec(
            args["refineexcludespec"], "Excluding voxels where ", "from refinement."
        )

    if args["offsetincludespec"] is not None:
        (
            args["offsetincludename"],
            args["offsetincludevals"],
        ) = tide_io.processnamespec(
            args["offsetincludespec"], "Including voxels where ", "in offset calculation."
        )

    if args["offsetexcludespec"] is not None:
        (
            args["offsetexcludename"],
            args["offsetexcludevals"],
        ) = tide_io.processnamespec(
            args["offsetexcludespec"], "Excluding voxels where ", "from offset calculation."
        )

    # motion processing
    if args["motionfilespec"] is not None:
        args["motionfilename"], args["motionfilecolspec"] = tide_io.parsefilespec(
            args["motionfilespec"]
        )
    else:
        args["motionfilename"] = None

    # process analysis modes
    if args["delaymapping"]:
        LGR.warning('Using "delaymapping" analysis mode. Overriding any affected arguments.')
        if args["passes"] == DEFAULT_PASSES:
            args["passes"] = DEFAULT_DELAYMAPPING_PASSES
        if args["despeckle_passes"] == DEFAULT_DESPECKLE_PASSES:
            args["despeckle_passes"] = DEFAULT_DELAYMAPPING_DESPECKLE_PASSES
        if args["gausssigma"] == DEFAULT_SPATIALFILT:
            args["gausssigma"] = DEFAULT_DELAYMAPPING_SPATIALFILT
        pf.setifnotset(args, "lagmin", DEFAULT_DELAYMAPPING_LAGMIN)
        pf.setifnotset(args, "lagmax", DEFAULT_DELAYMAPPING_LAGMAX)
        args["refineoffset"] = True
        args["refinedelay"] = True
        args["outputlevel"] = "normal"

    if args["denoising"]:
        LGR.warning('Using "denoising" analysis mode. Overriding any affected arguments.')
        if args["passes"] == DEFAULT_PASSES:
            args["passes"] = DEFAULT_DENOISING_PASSES
        if args["despeckle_passes"] == DEFAULT_DESPECKLE_PASSES:
            args["despeckle_passes"] = DEFAULT_DENOISING_DESPECKLE_PASSES
        if args["gausssigma"] == DEFAULT_SPATIALFILT:
            args["gausssigma"] = DEFAULT_DENOISING_SPATIALFILT
        if args["peakfittype"] == DEFAULT_PEAKFIT_TYPE:
            args["peakfittype"] = DEFAULT_DENOISING_PEAKFITTYPE
        pf.setifnotset(args, "lagmin", DEFAULT_DENOISING_LAGMIN)
        pf.setifnotset(args, "lagmax", DEFAULT_DENOISING_LAGMAX)
        args["refineoffset"] = True
        args["refinedelay"] = True
        args["zerooutbadfit"] = False
        args["dolinfitfilt"] = True

    if args["docvrmap"]:
        LGR.warning('Using "CVR" analysis mode. Overriding any affected arguments.')
        if args["regressorfile"] is None:
            raise ValueError(
                "CVR mapping requires an externally supplied regresssor file - terminating."
            )
        args["passvec"] = (
            DEFAULT_CVRMAPPING_FILTER_LOWERPASS,
            DEFAULT_CVRMAPPING_FILTER_UPPERPASS,
        )
        if args["despeckle_passes"] == DEFAULT_DESPECKLE_PASSES:
            args["despeckle_passes"] = DEFAULT_CVRMAPPING_DESPECKLE_PASSES
        if args["filterband"] == pf.DEFAULT_FILTERBAND:
            args["filterband"] = "None"
        pf.setifnotset(args, "lagmin", DEFAULT_CVRMAPPING_LAGMIN)
        pf.setifnotset(args, "lagmax", DEFAULT_CVRMAPPING_LAGMAX)
        args["preservefiltering"] = True
        args["passes"] = 1
        args["outputlevel"] = "min"
        args["dolinfitfilt"] = False

    if args["initregressorpreselect"]:
        LGR.warning('Using "globalpreselect" analysis mode. Overriding any affected arguments.')
        args["passes"] = 1
        args["despeckle_passes"] = 0
        args["refinedespeckled"] = False
        args["outputlevel"] = "normal"
        pf.setifnotset(args, "dolinfitfilt", False)
        args["saveintermediatemaps"] = False

    # configure the filter
    theobj, theprefilter = pf.postprocessfilteropts(Namespace(**args))
    args = vars(theobj)

    # process macros
    if args["venousrefine"]:
        LGR.warning('Using "venousrefine" macro. Overriding any affected arguments.')
        args["lagminthresh"] = 2.5
        args["lagmaxthresh"] = 6.0
        args["ampthresh"] = 0.5
        args["ampthreshfromsig"] = False
        args["lagmaskside"] = "upper"

    if args["nirs"]:
        LGR.warning('Using "nirs" macro. Overriding any affected arguments.')
        args["nothresh"] = True
        pf.setifnotset(args, "preservefiltering", False)
        args["dataiszeromean"] = True
        args["refineprenorm"] = "var"
        args["ampthresh"] = 0.7
        args["ampthreshfromsig"] = False
        args["lagminthresh"] = 0.1
        args["despeckle_passes"] = 0

    # process limitoutput
    if not args["limitoutput"]:
        args["outputlevel"] = "max"

    # output options
    if args["outputlevel"] == "min":
        args["saveconfoundfiltered"] = False
        args["savegaussout"] = False
        args["savecorrtimes"] = False
        args["savelagregressors"] = False
        args["savedespecklemasks"] = False
        args["saveminimumsLFOfiltfiles"] = False
        args["savenormalsLFOfiltfiles"] = False
        args["savemovingsignal"] = False
        args["saveallsLFOfiltfiles"] = False
    elif args["outputlevel"] == "less":
        args["saveconfoundfiltered"] = False
        args["savegaussout"] = False
        args["savecorrtimes"] = False
        args["savelagregressors"] = False
        args["savedespecklemasks"] = False
        args["saveminimumsLFOfiltfiles"] = True
        args["savenormalsLFOfiltfiles"] = False
        args["savemovingsignal"] = False
        args["saveallsLFOfiltfiles"] = False
    elif args["outputlevel"] == "normal":
        args["saveconfoundfiltered"] = False
        args["savegaussout"] = False
        args["savecorrtimes"] = False
        args["savelagregressors"] = False
        args["savedespecklemasks"] = False
        args["saveminimumsLFOfiltfiles"] = True
        args["savenormalsLFOfiltfiles"] = True
        args["savemovingsignal"] = False
        args["saveallsLFOfiltfiles"] = False
    elif args["outputlevel"] == "more":
        args["saveconfoundfiltered"] = False
        args["savegaussout"] = False
        args["savecorrtimes"] = False
        args["savelagregressors"] = True
        args["savedespecklemasks"] = False
        args["saveminimumsLFOfiltfiles"] = True
        args["savenormalsLFOfiltfiles"] = True
        args["savemovingsignal"] = True
        args["saveallsLFOfiltfiles"] = False
    elif args["outputlevel"] == "max":
        args["saveconfoundfiltered"] = True
        args["savegaussout"] = True
        args["savecorrtimes"] = True
        args["savelagregressors"] = True
        args["savedespecklemasks"] = True
        args["saveminimumsLFOfiltfiles"] = True
        args["savenormalsLFOfiltfiles"] = True
        args["savemovingsignal"] = True
        args["saveallsLFOfiltfiles"] = True
    else:
        print(f"illegal output level {args['outputlevel']}")
        sys.exit()

    # make the pass options dictionary
    args["passoptions"] = [
        {
            "similaritymetric": "riptide",
        },
        {
            "similaritymetric": "correlation",
        },
        {
            "similaritymetric": "correlation",
        },
    ]

    # dispersion calculation
    args["dispersioncalc_lower"] = args["lagmin"]
    args["dispersioncalc_upper"] = args["lagmax"]
    args["dispersioncalc_step"] = np.max(
        [
            (args["dispersioncalc_upper"] - args["dispersioncalc_lower"]) / 25,
            args["dispersioncalc_step"],
        ]
    )

    if args["territorymap"] is not None:
        (
            args["territorymapname"],
            args["territorymapincludevals"],
        ) = tide_io.processnamespec(
            args["territorymap"], "Including voxels where ", "in initial regressor calculation."
        )
    else:
        args["territorymapname"] = None
        args["territorymapincludevals"] = None

    # this is new enough to do retrospective regression filtering
    args["retroregresscompatible"] = True

    LGR.debug("\nafter postprocessing\n{}".format(args))

    # start the clock!
    tide_util.checkimports(args)

    return args, theprefilter



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import gc
import logging
import warnings
from typing import Any

import numpy as np
from numpy.typing import NDArray

import rapidtide.genericmultiproc as tide_genericmultiproc

warnings.simplefilter(action="ignore", category=FutureWarning)
LGR = logging.getLogger("GENERAL")


def _procOneVoxelCoherence(
    vox: int,
    voxelargs: list,
    **kwargs: Any,
) -> tuple[int, NDArray, NDArray, float, float]:
    """
    Process coherence for a single voxel.

    This function computes coherence values for a given voxel using the provided
    coherence calculator and fMRI time course data. It returns the voxel index
    along with coherence values and the location of the maximum coherence.

    Parameters
    ----------
    vox : int
        The voxel index being processed.
    voxelargs : list
        A list containing two elements: the coherence calculator object and
        the fMRI time course data (fmritc).
    **kwargs : Any
        Additional keyword arguments that can override default options:
        - alt : bool, optional (default: False)
            Flag to indicate alternative computation mode.
        - debug : bool, optional (default: False)
            Flag to enable debug printing.

    Returns
    -------
    tuple[int, NDArray, NDArray, float, float]
        A tuple containing:
        - vox : int
            The input voxel index
        - thecoherence_x : NDArray
            X-axis coherence values
        - thecoherence_y : NDArray
            Y-axis coherence values
        - thecoherence_y[maxindex] : float
            Maximum coherence value
        - thecoherence_x[maxindex] : float
            X-coordinate corresponding to maximum coherence

    Notes
    -----
    The function uses the `theCoherer.run()` method to compute coherence values.
    When `alt=True`, the function returns additional dummy values from the
    coherence calculation. The maximum coherence is determined using `np.argmax()`.

    Examples
    --------
    >>> result = _procOneVoxelCoherence(10, [coherer_obj, fmri_data], alt=True)
    >>> voxel_idx, x_vals, y_vals, max_coherence, max_x = result
    """
    options = {
        "alt": False,
        "debug": False,
    }
    options.update(kwargs)
    alt = options["alt"]
    debug = options["debug"]
    theCoherer, fmritc = voxelargs
    if debug:
        print(f"{alt=}")
    if alt:
        (
            thecoherence_y,
            thecoherence_x,
            globalmaxindex,
            dummy,
            dummy,
            dummy,
        ) = theCoherer.run(fmritc, trim=True, alt=True)
    else:
        thecoherence_y, thecoherence_x, globalmaxindex = theCoherer.run(fmritc, trim=True)
    maxindex = np.argmax(thecoherence_y)
    return (
        vox,
        thecoherence_x,
        thecoherence_y,
        thecoherence_y[maxindex],
        thecoherence_x[maxindex],
    )


def _packvoxeldata(voxnum: int, voxelargs: list) -> list:
    """
    Pack voxel data for processing.

    Parameters
    ----------
    voxnum : int
        The voxel number to extract from the second element of voxelargs.
    voxelargs : list
        A list containing voxel arguments where:
        - voxelargs[0] is the first voxel argument (returned as-is)
        - voxelargs[1] is a 2D array from which row voxnum is extracted

    Returns
    -------
    list
        A list containing:
        - voxelargs[0] (unchanged)
        - The voxnum-th row of voxelargs[1] as a 1D array

    Notes
    -----
    This function is typically used in voxel-based data processing workflows
    where data needs to be extracted and reorganized for further analysis.

    Examples
    --------
    >>> voxelargs = [10, [[1, 2, 3], [4, 5, 6], [7, 8, 9]]]
    >>> _packvoxeldata(1, voxelargs)
    [10, [4, 5, 6]]
    """
    return [
        voxelargs[0],
        (voxelargs[1])[voxnum, :],
    ]


def _unpackvoxeldata(retvals: tuple, voxelproducts: list) -> None:
    """
    Unpack voxel data from retvals tuple into corresponding voxel product lists.

    This function takes a tuple of voxel data and distributes the values into
    three separate voxel product lists based on the index specified in the first
    element of the retvals tuple.

    Parameters
    ----------
    retvals : tuple
        A tuple containing voxel data where:
        - retvals[0] : int, index for insertion
        - retvals[1] : unused
        - retvals[2] : value to insert into voxelproducts[0]
        - retvals[3] : value to insert into voxelproducts[1]
        - retvals[4] : value to insert into voxelproducts[2]
    voxelproducts : list
        A list of three voxel product arrays/lists where:
        - voxelproducts[0] : first voxel product array
        - voxelproducts[1] : second voxel product array
        - voxelproducts[2] : third voxel product array

    Returns
    -------
    None
        This function modifies the voxelproducts lists in-place and does not return anything.

    Notes
    -----
    The function assumes that retvals contains exactly 5 elements and that
    voxelproducts contains exactly 3 elements. The first element of retvals
    is used as an index to determine the position where values should be inserted
    into each of the three voxel product arrays.

    Examples
    --------
    >>> voxel1 = [0, 0, 0]
    >>> voxel2 = [0, 0, 0]
    >>> voxel3 = [0, 0, 0]
    >>> retvals = (1, None, 10, 20, 30)
    >>> voxelproducts = [voxel1, voxel2, voxel3]
    >>> _unpackvoxeldata(retvals, voxelproducts)
    >>> print(voxel1[1])
    10
    >>> print(voxel2[1])
    20
    >>> print(voxel3[1])
    30
    """
    (voxelproducts[0])[retvals[0]] = retvals[2]
    (voxelproducts[1])[retvals[0]] = retvals[3]
    (voxelproducts[2])[retvals[0]] = retvals[4]


def coherencepass(
    fmridata: NDArray,
    theCoherer: Any,
    coherencefunc: NDArray,
    coherencepeakval: NDArray,
    coherencepeakfreq: NDArray,
    alt: bool = False,
    chunksize: int = 1000,
    nprocs: int = 1,
    alwaysmultiproc: bool = False,
    showprogressbar: bool = True,
    debug: bool = False,
) -> int:
    """
    Perform coherence analysis on fMRI data across voxels using multiprocessing.

    This function applies coherence analysis to each voxel in the input fMRI data,
    storing results in the provided output arrays. It supports parallel processing
    for improved performance and includes optional debugging and progress tracking.

    Parameters
    ----------
    fmridata : NDArray
        Input fMRI data array with shape (time, voxels).
    theCoherer : Any
        Object or function used to perform coherence calculations.
    coherencefunc : NDArray
        Array to store coherence function results for each voxel.
    coherencepeakval : NDArray
        Array to store peak coherence values for each voxel.
    coherencepeakfreq : NDArray
        Array to store peak coherence frequencies for each voxel.
    alt : bool, optional
        If True, use alternative coherence calculation method. Default is False.
    chunksize : int, optional
        Number of voxels to process in each chunk during multiprocessing.
        Default is 1000.
    nprocs : int, optional
        Number of processes to use for multiprocessing. Default is 1.
    alwaysmultiproc : bool, optional
        If True, always use multiprocessing even for small datasets.
        Default is False.
    showprogressbar : bool, optional
        If True, display a progress bar during processing. Default is True.
    debug : bool, optional
        If True, enable debug logging. Default is False.

    Returns
    -------
    int
        Total number of voxels processed.

    Notes
    -----
    This function uses `tide_genericmultiproc.run_multiproc` to distribute
    voxel-wise coherence computations across multiple processes. The results
    are stored directly into the provided output arrays (`coherencefunc`,
    `coherencepeakval`, `coherencepeakfreq`).

    Examples
    --------
    >>> import numpy as np
    >>> fmri_data = np.random.rand(100, 50)
    >>> coherer = SomeCohererClass()
    >>> coherence_func = np.zeros((100, 50))
    >>> peak_val = np.zeros((1, 50))
    >>> peak_freq = np.zeros((1, 50))
    >>> n_voxels = coherencepass(
    ...     fmri_data, coherer, coherence_func, peak_val, peak_freq
    ... )
    >>> print(f"Processed {n_voxels} voxels")
    """
    inputshape = np.shape(fmridata)
    voxelargs = [theCoherer, fmridata]
    voxelfunc = _procOneVoxelCoherence
    packfunc = _packvoxeldata
    unpackfunc = _unpackvoxeldata
    voxeltargets = [coherencefunc, coherencepeakval, coherencepeakfreq]
    voxelmask = np.ones_like(fmridata[:, 0])

    volumetotal = tide_genericmultiproc.run_multiproc(
        voxelfunc,
        packfunc,
        unpackfunc,
        voxelargs,
        voxeltargets,
        inputshape,
        voxelmask,
        LGR,
        nprocs,
        alwaysmultiproc,
        showprogressbar,
        chunksize,
        alt=alt,
        debug=debug,
    )
    LGR.info(f"\nCoherence performed on {volumetotal} voxels")

    # garbage collect
    uncollected = gc.collect()
    if uncollected != 0:
        LGR.info(f"garbage collected - unable to collect {uncollected} objects")
    else:
        LGR.info("garbage collected")

    return volumetotal



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
import tempfile
import time
from unittest.mock import MagicMock, patch

import numpy as np
import pytest

from rapidtide.workflows.calctexticc import (
    _get_parser,
    calctexticc,
    makdcommandlinelist,
    parsetextmeasurementlist,
)


def _capture_writenpvecs():
    """Return a side_effect function and its captured data list.

    calctexticc reuses a single numpy array for all writenpvecs calls,
    so the mock records a reference to the same mutating array.  This
    helper copies data at call time so each captured entry is independent.
    """
    captured = []

    def _side_effect(data, filename, **kwargs):
        captured.append((data.copy(), filename))

    return _side_effect, captured


# ---- _get_parser tests ----


def test_get_parser_returns_parser(debug=False):
    """Test that _get_parser returns an ArgumentParser."""
    parser = _get_parser()
    assert isinstance(parser, argparse.ArgumentParser)
    if debug:
        print("Parser created successfully")


def test_get_parser_required_args(debug=False):
    """Test that parser requires the three positional arguments."""
    parser = _get_parser()
    # Should fail with no arguments
    with pytest.raises(SystemExit):
        parser.parse_args([])


def test_get_parser_with_valid_args(debug=False):
    """Test parser with valid positional arguments using a real temp file for measurementlist."""
    with tempfile.NamedTemporaryFile(mode="w", suffix=".txt", delete=False) as f:
        f.write("0 1\n2 3\n")
        measfile = f.name

    parser = _get_parser()
    args = parser.parse_args(["data1.txt,data2.txt", measfile, "output_root"])
    assert args.datafile == "data1.txt,data2.txt"
    assert args.outputroot == "output_root"
    assert args.demedian is False
    assert args.demean is False
    assert args.nocache is False
    assert args.debug is False
    assert args.deepdebug is False

    import os

    os.unlink(measfile)

    if debug:
        print(f"Parsed args: {args}")


def test_get_parser_optional_flags(debug=False):
    """Test parser with all optional flags enabled."""
    with tempfile.NamedTemporaryFile(mode="w", suffix=".txt", delete=False) as f:
        f.write("0 1\n")
        measfile = f.name

    parser = _get_parser()
    args = parser.parse_args(
        [
            "data.txt",
            measfile,
            "out",
            "--demedian",
            "--demean",
            "--nocache",
            "--debug",
            "--deepdebug",
        ]
    )
    assert args.demedian is True
    assert args.demean is True
    assert args.nocache is True
    assert args.debug is True
    assert args.deepdebug is True

    import os

    os.unlink(measfile)


# ---- parsetextmeasurementlist tests ----


def test_parsetextmeasurementlist_single_values(debug=False):
    """Test parsing measurement list with single values (volume only, file=0)."""
    measlist = np.array([["0", "1"], ["2", "3"]])
    numfiles = 1
    filesel, volumesel = parsetextmeasurementlist(measlist, numfiles)

    assert filesel.shape == (2, 2)
    assert volumesel.shape == (2, 2)
    # Single values => file index defaults to 0
    np.testing.assert_array_equal(filesel, np.zeros((2, 2), dtype=int))
    np.testing.assert_array_equal(volumesel, np.array([[0, 1], [2, 3]]))

    if debug:
        print(f"filesel:\n{filesel}")
        print(f"volumesel:\n{volumesel}")


def test_parsetextmeasurementlist_comma_separated(debug=False):
    """Test parsing measurement list with file,volume entries."""
    measlist = np.array([["0,1", "1,2"], ["0,3", "2,4"]])
    numfiles = 3
    filesel, volumesel = parsetextmeasurementlist(measlist, numfiles)

    expected_filesel = np.array([[0, 1], [0, 2]])
    expected_volumesel = np.array([[1, 2], [3, 4]])
    np.testing.assert_array_equal(filesel, expected_filesel)
    np.testing.assert_array_equal(volumesel, expected_volumesel)

    if debug:
        print(f"filesel:\n{filesel}")
        print(f"volumesel:\n{volumesel}")


def test_parsetextmeasurementlist_mixed_formats(debug=False):
    """Test parsing measurement list with mixed single and comma-separated entries."""
    measlist = np.array([["0,5", "3"], ["1,2", "0,7"]])
    numfiles = 2
    filesel, volumesel = parsetextmeasurementlist(measlist, numfiles)

    expected_filesel = np.array([[0, 0], [1, 0]])
    expected_volumesel = np.array([[5, 3], [2, 7]])
    np.testing.assert_array_equal(filesel, expected_filesel)
    np.testing.assert_array_equal(volumesel, expected_volumesel)


def test_parsetextmeasurementlist_debug_output(debug=False):
    """Test that debug mode prints without errors."""
    measlist = np.array([["0,1", "2"]])
    numfiles = 1
    filesel, volumesel = parsetextmeasurementlist(measlist, numfiles, debug=True)
    assert filesel.shape == (1, 2)


def test_parsetextmeasurementlist_too_many_commas(debug=False):
    """Test that entries with more than one comma cause sys.exit."""
    measlist = np.array([["0,1,2"]])
    numfiles = 3
    with pytest.raises(SystemExit):
        parsetextmeasurementlist(measlist, numfiles)


def test_parsetextmeasurementlist_file_out_of_range(debug=False):
    """Test that file index exceeding numfiles causes sys.exit."""
    measlist = np.array([["5,0"]])
    numfiles = 3  # valid file indices: 0, 1, 2
    with pytest.raises(SystemExit):
        parsetextmeasurementlist(measlist, numfiles)


def test_parsetextmeasurementlist_single_element(debug=False):
    """Test with a single element measurement list."""
    measlist = np.array([["0,0"]])
    numfiles = 1
    filesel, volumesel = parsetextmeasurementlist(measlist, numfiles)
    assert filesel.shape == (1, 1)
    assert volumesel.shape == (1, 1)
    assert filesel[0, 0] == 0
    assert volumesel[0, 0] == 0


# ---- makdcommandlinelist tests ----


def test_makdcommandlinelist_without_extra(debug=False):
    """Test command line list generation without extra info."""
    arglist = ["python", "calctexticc", "--debug", "data.txt"]
    starttime = time.time() - 5.0
    endtime = time.time()

    with patch("rapidtide.workflows.calctexticc.tide_util.version") as mock_version:
        mock_version.return_value = ("v2.9.0", "abc123", "2024-01-01", False)
        result = makdcommandlinelist(arglist, starttime, endtime)

    assert len(result) == 4
    assert result[0].startswith("# Processed on")
    assert "Processing took" in result[1]
    assert "v2.9.0" in result[2]
    assert result[3] == "python calctexticc --debug data.txt"

    if debug:
        for line in result:
            print(line)


def test_makdcommandlinelist_with_extra(debug=False):
    """Test command line list generation with extra info."""
    arglist = ["python", "calctexticc"]
    starttime = time.time() - 2.0
    endtime = time.time()

    with patch("rapidtide.workflows.calctexticc.tide_util.version") as mock_version:
        mock_version.return_value = ("v2.9.0", "abc123", "2024-01-01", False)
        result = makdcommandlinelist(arglist, starttime, endtime, extra="ICC time: 1.5 ms")

    assert len(result) == 5
    assert result[0].startswith("# Processed on")
    assert "Processing took" in result[1]
    assert "v2.9.0" in result[2]
    assert result[3] == "# ICC time: 1.5 ms"
    assert result[4] == "python calctexticc"

    if debug:
        for line in result:
            print(line)


def test_makdcommandlinelist_timing(debug=False):
    """Test that timing information is computed correctly."""
    arglist = ["test"]
    starttime = 1000.0
    endtime = 1010.5

    with patch("rapidtide.workflows.calctexticc.tide_util.version") as mock_version:
        mock_version.return_value = ("v1.0", "sha", "date", False)
        result = makdcommandlinelist(arglist, starttime, endtime)

    assert "10.500" in result[1]


def test_makdcommandlinelist_empty_arglist(debug=False):
    """Test with empty argument list."""
    arglist = []
    starttime = time.time()
    endtime = starttime + 1.0

    with patch("rapidtide.workflows.calctexticc.tide_util.version") as mock_version:
        mock_version.return_value = ("v1.0", "sha", "date", False)
        result = makdcommandlinelist(arglist, starttime, endtime)

    # Command line should be empty string
    assert result[-1] == ""


# ---- calctexticc tests ----


def _make_icc_test_data(numsubjs=4, nummeas=3, numvals=5, rng=None):
    """Helper to create synthetic ICC test data.

    Creates data where each subject has a consistent pattern across measurements
    plus some noise, so ICC should be moderately high.
    """
    if rng is None:
        rng = np.random.RandomState(42)

    # Subject effects (consistent across measurements)
    subject_effects = rng.randn(numsubjs, numvals) * 5.0
    # Measurement noise
    noise = rng.randn(numsubjs, nummeas, numvals) * 1.0

    # data[val, subj*nummeas + meas] format
    data = np.zeros((numvals, numsubjs * nummeas))
    for subj in range(numsubjs):
        for meas in range(nummeas):
            data[:, subj * nummeas + meas] = subject_effects[subj, :] + noise[subj, meas, :]

    return data


def _make_measlist(numsubjs, nummeas):
    """Helper to create a simple measurement list for a single data file."""
    measlist = np.zeros((nummeas, numsubjs), dtype="U10")
    for subj in range(numsubjs):
        for meas in range(nummeas):
            measlist[meas, subj] = str(subj * nummeas + meas)
    return measlist


def test_calctexticc_basic(debug=False):
    """Test basic ICC calculation with synthetic data."""
    rng = np.random.RandomState(42)
    numsubjs = 4
    nummeas = 2
    numvals = 5

    data = _make_icc_test_data(numsubjs=numsubjs, nummeas=nummeas, numvals=numvals, rng=rng)
    measlist = _make_measlist(numsubjs, nummeas)

    side_effect, captured = _capture_writenpvecs()

    with tempfile.TemporaryDirectory() as tmpdir:
        outputroot = f"{tmpdir}/test_icc"

        args = argparse.Namespace(
            datafile="dummy.txt",
            measurementlist="dummy_meas.txt",
            outputroot=outputroot,
            debug=debug,
            deepdebug=False,
            demedian=False,
            demean=False,
            nocache=False,
        )

        with (
            patch("rapidtide.workflows.calctexticc.tide_io.readvecs") as mock_readvecs,
            patch("rapidtide.workflows.calctexticc.tide_io.writenpvecs", side_effect=side_effect),
            patch("rapidtide.workflows.calctexticc.tide_io.writevec"),
            patch("rapidtide.workflows.calctexticc.tide_util.version") as mock_version,
        ):
            mock_version.return_value = ("v2.9.0", "abc123", "2024-01-01", False)
            mock_readvecs.side_effect = [measlist, data, data]

            calctexticc(args)

            # Should have written 4 output files
            assert len(captured) == 4

            # Check ICC output shape
            icc_values, icc_filename = captured[0]
            assert icc_values.shape == (numvals,)

            # Check filenames
            assert icc_filename == f"{outputroot}_ICC.txt"
            assert captured[1][1] == f"{outputroot}_r_var.txt"
            assert captured[2][1] == f"{outputroot}_e_var.txt"
            assert captured[3][1] == f"{outputroot}_session_effect_F.txt"

        if debug:
            print(f"ICC values: {icc_values}")


def test_calctexticc_demedian(debug=False):
    """Test ICC calculation with demedian option."""
    rng = np.random.RandomState(99)
    numsubjs = 3
    nummeas = 2
    numvals = 4

    data = _make_icc_test_data(numsubjs=numsubjs, nummeas=nummeas, numvals=numvals, rng=rng)
    data += 100.0
    measlist = _make_measlist(numsubjs, nummeas)

    side_effect, captured = _capture_writenpvecs()

    with tempfile.TemporaryDirectory() as tmpdir:
        outputroot = f"{tmpdir}/test_demedian"

        args = argparse.Namespace(
            datafile="dummy.txt",
            measurementlist="dummy_meas.txt",
            outputroot=outputroot,
            debug=debug,
            deepdebug=False,
            demedian=True,
            demean=False,
            nocache=False,
        )

        with (
            patch("rapidtide.workflows.calctexticc.tide_io.readvecs") as mock_readvecs,
            patch("rapidtide.workflows.calctexticc.tide_io.writenpvecs", side_effect=side_effect),
            patch("rapidtide.workflows.calctexticc.tide_io.writevec"),
            patch("rapidtide.workflows.calctexticc.tide_util.version") as mock_version,
        ):
            mock_version.return_value = ("v2.9.0", "abc123", "2024-01-01", False)
            mock_readvecs.side_effect = [measlist, data, data]

            calctexticc(args)

            assert len(captured) == 4
            icc_values = captured[0][0]
            assert icc_values.shape == (numvals,)

            if debug:
                print(f"Demedian ICC values: {icc_values}")


def test_calctexticc_demean(debug=False):
    """Test ICC calculation with demean option."""
    rng = np.random.RandomState(77)
    numsubjs = 3
    nummeas = 2
    numvals = 4

    data = _make_icc_test_data(numsubjs=numsubjs, nummeas=nummeas, numvals=numvals, rng=rng)
    data += 50.0
    measlist = _make_measlist(numsubjs, nummeas)

    side_effect, captured = _capture_writenpvecs()

    with tempfile.TemporaryDirectory() as tmpdir:
        outputroot = f"{tmpdir}/test_demean"

        args = argparse.Namespace(
            datafile="dummy.txt",
            measurementlist="dummy_meas.txt",
            outputroot=outputroot,
            debug=debug,
            deepdebug=False,
            demedian=False,
            demean=True,
            nocache=False,
        )

        with (
            patch("rapidtide.workflows.calctexticc.tide_io.readvecs") as mock_readvecs,
            patch("rapidtide.workflows.calctexticc.tide_io.writenpvecs", side_effect=side_effect),
            patch("rapidtide.workflows.calctexticc.tide_io.writevec"),
            patch("rapidtide.workflows.calctexticc.tide_util.version") as mock_version,
        ):
            mock_version.return_value = ("v2.9.0", "abc123", "2024-01-01", False)
            mock_readvecs.side_effect = [measlist, data, data]

            calctexticc(args)

            assert len(captured) == 4
            icc_values = captured[0][0]
            assert icc_values.shape == (numvals,)

            if debug:
                print(f"Demean ICC values: {icc_values}")


def test_calctexticc_nocache(debug=False):
    """Test ICC calculation with nocache option."""
    rng = np.random.RandomState(55)
    numsubjs = 3
    nummeas = 2
    numvals = 3

    data = _make_icc_test_data(numsubjs=numsubjs, nummeas=nummeas, numvals=numvals, rng=rng)
    measlist = _make_measlist(numsubjs, nummeas)

    side_effect, captured = _capture_writenpvecs()

    with tempfile.TemporaryDirectory() as tmpdir:
        outputroot = f"{tmpdir}/test_nocache"

        args = argparse.Namespace(
            datafile="dummy.txt",
            measurementlist="dummy_meas.txt",
            outputroot=outputroot,
            debug=debug,
            deepdebug=False,
            demedian=False,
            demean=False,
            nocache=True,
        )

        with (
            patch("rapidtide.workflows.calctexticc.tide_io.readvecs") as mock_readvecs,
            patch("rapidtide.workflows.calctexticc.tide_io.writenpvecs", side_effect=side_effect),
            patch("rapidtide.workflows.calctexticc.tide_io.writevec"),
            patch("rapidtide.workflows.calctexticc.tide_util.version") as mock_version,
        ):
            mock_version.return_value = ("v2.9.0", "abc123", "2024-01-01", False)
            mock_readvecs.side_effect = [measlist, data, data]

            calctexticc(args)

            assert len(captured) == 4
            icc_values = captured[0][0]
            assert icc_values.shape == (numvals,)


def test_calctexticc_multiple_files(debug=False):
    """Test ICC calculation with data spread across multiple files."""
    rng = np.random.RandomState(33)
    numsubjs = 3
    nummeas = 2
    numvals = 4

    # Create two data files, each with a subset of volumes
    data_file0 = rng.randn(numvals, numsubjs)  # 3 volumes in file 0
    data_file1 = rng.randn(numvals, numsubjs)  # 3 volumes in file 1

    # Measurement list uses file,volume format
    measlist = np.array(
        [["0,0", "0,1", "0,2"], ["1,0", "1,1", "1,2"]],  # meas 0: all from file 0
        dtype="U10",  # meas 1: all from file 1
    )

    side_effect, captured = _capture_writenpvecs()

    with tempfile.TemporaryDirectory() as tmpdir:
        outputroot = f"{tmpdir}/test_multifile"

        args = argparse.Namespace(
            datafile="file0.txt,file1.txt",
            measurementlist="dummy_meas.txt",
            outputroot=outputroot,
            debug=debug,
            deepdebug=False,
            demedian=False,
            demean=False,
            nocache=False,
        )

        with (
            patch("rapidtide.workflows.calctexticc.tide_io.readvecs") as mock_readvecs,
            patch("rapidtide.workflows.calctexticc.tide_io.writenpvecs", side_effect=side_effect),
            patch("rapidtide.workflows.calctexticc.tide_io.writevec"),
            patch("rapidtide.workflows.calctexticc.tide_util.version") as mock_version,
        ):
            mock_version.return_value = ("v2.9.0", "abc123", "2024-01-01", False)
            # Calls: measlist, header0, header1, data0, data1
            mock_readvecs.side_effect = [
                measlist,
                data_file0,
                data_file1,
                data_file0,
                data_file1,
            ]

            calctexticc(args)

            assert len(captured) == 4


def test_calctexticc_deepdebug(debug=False):
    """Test ICC calculation with deepdebug output."""
    rng = np.random.RandomState(11)
    numsubjs = 2
    nummeas = 2
    numvals = 2

    data = _make_icc_test_data(numsubjs=numsubjs, nummeas=nummeas, numvals=numvals, rng=rng)
    measlist = _make_measlist(numsubjs, nummeas)

    with tempfile.TemporaryDirectory() as tmpdir:
        outputroot = f"{tmpdir}/test_deepdebug"

        args = argparse.Namespace(
            datafile="dummy.txt",
            measurementlist="dummy_meas.txt",
            outputroot=outputroot,
            debug=True,
            deepdebug=True,
            demedian=False,
            demean=False,
            nocache=False,
        )

        with (
            patch("rapidtide.workflows.calctexticc.tide_io.readvecs") as mock_readvecs,
            patch("rapidtide.workflows.calctexticc.tide_io.writenpvecs"),
            patch("rapidtide.workflows.calctexticc.tide_io.writevec"),
            patch("rapidtide.workflows.calctexticc.tide_util.version") as mock_version,
        ):
            mock_version.return_value = ("v2.9.0", "abc123", "2024-01-01", False)
            mock_readvecs.side_effect = [measlist, data, data]

            # Should not raise even with deepdebug enabled
            calctexticc(args)


def test_calctexticc_icc_values_correct(debug=False):
    """Test that ICC values are high for nearly-perfectly reproducible data."""
    rng = np.random.RandomState(42)
    numsubjs = 6
    nummeas = 4
    numvals = 3

    # Create nearly-perfect data: strong subject effects with tiny noise
    data = np.zeros((numvals, numsubjs * nummeas))
    subject_values = np.linspace(1.0, 10.0, numsubjs)
    for subj in range(numsubjs):
        for meas in range(nummeas):
            for v in range(numvals):
                data[v, subj * nummeas + meas] = (
                    subject_values[subj] * (v + 1) + rng.randn() * 0.01
                )

    measlist = _make_measlist(numsubjs, nummeas)
    side_effect, captured = _capture_writenpvecs()

    with tempfile.TemporaryDirectory() as tmpdir:
        outputroot = f"{tmpdir}/test_high_icc"

        args = argparse.Namespace(
            datafile="dummy.txt",
            measurementlist="dummy_meas.txt",
            outputroot=outputroot,
            debug=debug,
            deepdebug=False,
            demedian=False,
            demean=False,
            nocache=True,
        )

        with (
            patch("rapidtide.workflows.calctexticc.tide_io.readvecs") as mock_readvecs,
            patch("rapidtide.workflows.calctexticc.tide_io.writenpvecs", side_effect=side_effect),
            patch("rapidtide.workflows.calctexticc.tide_io.writevec"),
            patch("rapidtide.workflows.calctexticc.tide_util.version") as mock_version,
        ):
            mock_version.return_value = ("v2.9.0", "abc123", "2024-01-01", False)
            mock_readvecs.side_effect = [measlist, data, data]

            calctexticc(args)

            icc_values = captured[0][0]  # ICC
            e_var_values = captured[2][0]  # e_var

            # Near-perfect reproducibility should give ICC close to 1.0
            assert np.all(icc_values > 0.99), f"Expected ICC > 0.99, got {icc_values}"

            # Error variance should be very small
            assert np.all(e_var_values < 0.01), f"Expected small e_var, got {e_var_values}"

            if debug:
                print(f"High ICC values: {icc_values}")
                print(f"Error variance: {e_var_values}")


def test_calctexticc_zero_icc_data(debug=False):
    """Test ICC calculation with pure noise data (ICC should be near 0)."""
    rng = np.random.RandomState(123)
    numsubjs = 10
    nummeas = 3
    numvals = 3

    # Pure noise - no subject effects, ICC should be near 0
    data = rng.randn(numvals, numsubjs * nummeas)
    measlist = _make_measlist(numsubjs, nummeas)

    side_effect, captured = _capture_writenpvecs()

    with tempfile.TemporaryDirectory() as tmpdir:
        outputroot = f"{tmpdir}/test_noise"

        args = argparse.Namespace(
            datafile="dummy.txt",
            measurementlist="dummy_meas.txt",
            outputroot=outputroot,
            debug=debug,
            deepdebug=False,
            demedian=False,
            demean=False,
            nocache=True,
        )

        with (
            patch("rapidtide.workflows.calctexticc.tide_io.readvecs") as mock_readvecs,
            patch("rapidtide.workflows.calctexticc.tide_io.writenpvecs", side_effect=side_effect),
            patch("rapidtide.workflows.calctexticc.tide_io.writevec"),
            patch("rapidtide.workflows.calctexticc.tide_util.version") as mock_version,
        ):
            mock_version.return_value = ("v2.9.0", "abc123", "2024-01-01", False)
            mock_readvecs.side_effect = [measlist, data, data]

            calctexticc(args)

            icc_values = captured[0][0]
            # For pure noise, ICC should be close to 0 (but can be slightly negative)
            assert np.all(
                np.abs(icc_values) < 0.5
            ), f"Expected low ICC for noise, got {icc_values}"

            if debug:
                print(f"Noise ICC values: {icc_values}")


def test_calctexticc_output_variances(debug=False):
    """Test that e_var outputs are non-negative for well-behaved data."""
    rng = np.random.RandomState(88)
    numsubjs = 5
    nummeas = 3
    numvals = 4

    data = _make_icc_test_data(numsubjs=numsubjs, nummeas=nummeas, numvals=numvals, rng=rng)
    measlist = _make_measlist(numsubjs, nummeas)

    side_effect, captured = _capture_writenpvecs()

    with tempfile.TemporaryDirectory() as tmpdir:
        outputroot = f"{tmpdir}/test_var"

        args = argparse.Namespace(
            datafile="dummy.txt",
            measurementlist="dummy_meas.txt",
            outputroot=outputroot,
            debug=debug,
            deepdebug=False,
            demedian=False,
            demean=False,
            nocache=False,
        )

        with (
            patch("rapidtide.workflows.calctexticc.tide_io.readvecs") as mock_readvecs,
            patch("rapidtide.workflows.calctexticc.tide_io.writenpvecs", side_effect=side_effect),
            patch("rapidtide.workflows.calctexticc.tide_io.writevec"),
            patch("rapidtide.workflows.calctexticc.tide_util.version") as mock_version,
        ):
            mock_version.return_value = ("v2.9.0", "abc123", "2024-01-01", False)
            mock_readvecs.side_effect = [measlist, data, data]

            calctexticc(args)

            e_var_values = captured[2][0]
            session_f_values = captured[3][0]

            # Error variance should always be non-negative (it's MSE)
            assert np.all(e_var_values >= 0), f"e_var should be non-negative: {e_var_values}"
            # Session effect F should be non-negative
            assert np.all(session_f_values >= 0), f"F should be non-negative: {session_f_values}"

            if debug:
                print(f"e_var: {e_var_values}")
                print(f"session_F: {session_f_values}")


# ---- main test entry point ----


def test_calctexticc(debug=False):
    """Run all calctexticc sub-tests."""
    # _get_parser tests
    test_get_parser_returns_parser(debug=debug)
    test_get_parser_required_args(debug=debug)
    test_get_parser_with_valid_args(debug=debug)
    test_get_parser_optional_flags(debug=debug)

    # parsetextmeasurementlist tests
    test_parsetextmeasurementlist_single_values(debug=debug)
    test_parsetextmeasurementlist_comma_separated(debug=debug)
    test_parsetextmeasurementlist_mixed_formats(debug=debug)
    test_parsetextmeasurementlist_debug_output(debug=debug)
    test_parsetextmeasurementlist_too_many_commas(debug=debug)
    test_parsetextmeasurementlist_file_out_of_range(debug=debug)
    test_parsetextmeasurementlist_single_element(debug=debug)

    # makdcommandlinelist tests
    test_makdcommandlinelist_without_extra(debug=debug)
    test_makdcommandlinelist_with_extra(debug=debug)
    test_makdcommandlinelist_timing(debug=debug)
    test_makdcommandlinelist_empty_arglist(debug=debug)

    # calctexticc tests
    test_calctexticc_basic(debug=debug)
    test_calctexticc_demedian(debug=debug)
    test_calctexticc_demean(debug=debug)
    test_calctexticc_nocache(debug=debug)
    test_calctexticc_multiple_files(debug=debug)
    test_calctexticc_deepdebug(debug=debug)
    test_calctexticc_icc_values_correct(debug=debug)
    test_calctexticc_zero_icc_data(debug=debug)
    test_calctexticc_output_variances(debug=debug)


if __name__ == "__main__":
    test_calctexticc(debug=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import os
import tempfile

import nibabel as nib
import numpy as np

import rapidtide.filter as tide_filt
import rapidtide.workflows.localflow as lf
import rapidtide.workflows.parser_funcs as pf

# ==================== Helpers ====================


def _make_broadband_signal(timepoints, Fs, delay=0.0, seed=42):
    """Generate a broadband signal as a sum of sinusoids."""
    rng = np.random.RandomState(seed)
    t = np.arange(timepoints) / Fs
    signal = np.zeros(timepoints, dtype=float)
    for _ in range(30):
        freq = rng.uniform(0.01, 0.15)
        phase = rng.uniform(0, 2 * np.pi)
        signal += np.sin(2 * np.pi * freq * (t - delay) + phase)
    return signal


def _make_4d_nifti(tmpdir, shape=(6, 6, 4, 50), tr=2.0, name="test4d"):
    """Create a 4D NIfTI file with correlated broadband signals and return path."""
    nx, ny, nz, nt = shape
    Fs = 1.0 / tr
    rng = np.random.RandomState(42)
    data = np.zeros(shape, dtype=np.float64)

    # Fill with broadband signals + noise so that mask and correlations are meaningful
    base_signal = _make_broadband_signal(nt, Fs, delay=0.0, seed=42)
    for x in range(nx):
        for y in range(ny):
            for z in range(nz):
                delay = 0.1 * (x + y + z)
                sig = _make_broadband_signal(nt, Fs, delay=delay, seed=x * 100 + y * 10 + z)
                data[x, y, z, :] = sig + 0.5 * rng.randn(nt)

    # Add a base offset so the mask works (makemask uses threshpct)
    data += 100.0

    img = nib.Nifti1Image(data, affine=np.diag([2.0, 2.0, 2.0, 1.0]))
    hdr = img.header
    hdr.set_zooms((2.0, 2.0, 2.0, tr))
    hdr.set_xyzt_units("mm", "sec")
    filepath = os.path.join(tmpdir, name + ".nii.gz")
    nib.save(img, filepath)
    return filepath, data


# ==================== xyz2index tests ====================


def xyz2index_basic(debug=False):
    """Test xyz2index with basic coordinates."""
    if debug:
        print("xyz2index_basic")
    # For a (10, 10, 10) array, index = z + y*zsize + x*zsize*ysize
    result = lf.xyz2index(0, 0, 0, 10, 10, 10)
    assert result == 0

    result = lf.xyz2index(0, 0, 1, 10, 10, 10)
    assert result == 1

    result = lf.xyz2index(0, 1, 0, 10, 10, 10)
    assert result == 10

    result = lf.xyz2index(1, 0, 0, 10, 10, 10)
    assert result == 100


def xyz2index_corners(debug=False):
    """Test xyz2index with corner coordinates."""
    if debug:
        print("xyz2index_corners")
    xsize, ysize, zsize = 5, 7, 3
    # Last valid index
    result = lf.xyz2index(4, 6, 2, xsize, ysize, zsize)
    assert result == 2 + 6 * 3 + 4 * 3 * 7
    assert result == xsize * ysize * zsize - 1


def xyz2index_out_of_bounds(debug=False):
    """Test xyz2index returns -1 for out-of-bounds coordinates."""
    if debug:
        print("xyz2index_out_of_bounds")
    assert lf.xyz2index(-1, 0, 0, 10, 10, 10) == -1
    assert lf.xyz2index(0, -1, 0, 10, 10, 10) == -1
    assert lf.xyz2index(0, 0, -1, 10, 10, 10) == -1
    assert lf.xyz2index(10, 0, 0, 10, 10, 10) == -1
    assert lf.xyz2index(0, 10, 0, 10, 10, 10) == -1
    assert lf.xyz2index(0, 0, 10, 10, 10, 10) == -1


def xyz2index_boundary(debug=False):
    """Test xyz2index at boundary values."""
    if debug:
        print("xyz2index_boundary")
    # Last valid coordinate on each axis
    assert lf.xyz2index(9, 0, 0, 10, 10, 10) >= 0
    assert lf.xyz2index(0, 9, 0, 10, 10, 10) >= 0
    assert lf.xyz2index(0, 0, 9, 10, 10, 10) >= 0
    # One past the end
    assert lf.xyz2index(10, 0, 0, 10, 10, 10) == -1
    assert lf.xyz2index(0, 10, 0, 10, 10, 10) == -1
    assert lf.xyz2index(0, 0, 10, 10, 10, 10) == -1


def xyz2index_nonsquare(debug=False):
    """Test xyz2index with non-square dimensions."""
    if debug:
        print("xyz2index_nonsquare")
    xsize, ysize, zsize = 3, 5, 7
    result = lf.xyz2index(2, 4, 6, xsize, ysize, zsize)
    expected = 6 + 4 * 7 + 2 * 7 * 5
    assert result == expected


# ==================== index2xyz tests ====================


def index2xyz_basic(debug=False):
    """Test index2xyz with basic indices."""
    if debug:
        print("index2xyz_basic")
    x, y, z = lf.index2xyz(0, 10, 10)
    assert (x, y, z) == (0, 0, 0)

    x, y, z = lf.index2xyz(1, 10, 10)
    assert (x, y, z) == (0, 0, 1)

    x, y, z = lf.index2xyz(10, 10, 10)
    assert (x, y, z) == (0, 1, 0)

    x, y, z = lf.index2xyz(100, 10, 10)
    assert (x, y, z) == (1, 0, 0)


def index2xyz_nonsquare(debug=False):
    """Test index2xyz with non-square dimensions."""
    if debug:
        print("index2xyz_nonsquare")
    ysize, zsize = 5, 7
    # index = z + y*zsize + x*zsize*ysize
    idx = 6 + 4 * 7 + 2 * 7 * 5
    x, y, z = lf.index2xyz(idx, ysize, zsize)
    assert (x, y, z) == (2, 4, 6)


def index2xyz_roundtrip(debug=False):
    """Test that xyz2index and index2xyz are inverses."""
    if debug:
        print("index2xyz_roundtrip")
    xsize, ysize, zsize = 4, 5, 6
    for x in range(xsize):
        for y in range(ysize):
            for z in range(zsize):
                idx = lf.xyz2index(x, y, z, xsize, ysize, zsize)
                x2, y2, z2 = lf.index2xyz(idx, ysize, zsize)
                assert (x2, y2, z2) == (x, y, z), f"Roundtrip failed for ({x},{y},{z})"


def index2xyz_sequential(debug=False):
    """Test index2xyz produces sequential coordinates."""
    if debug:
        print("index2xyz_sequential")
    ysize, zsize = 3, 4
    # Check that indices 0..xsize*ysize*zsize-1 produce all valid coordinates
    xsize = 2
    coords = set()
    for idx in range(xsize * ysize * zsize):
        x, y, z = lf.index2xyz(idx, ysize, zsize)
        assert 0 <= x < xsize
        assert 0 <= y < ysize
        assert 0 <= z < zsize
        coords.add((x, y, z))
    assert len(coords) == xsize * ysize * zsize


# ==================== getcorrloc tests ====================


def getcorrloc_identical_signals(debug=False):
    """Test getcorrloc with identical signals returns high correlation at lag 0."""
    if debug:
        print("getcorrloc_identical_signals")
    Fs = 10.0
    npts = 500
    signal = _make_broadband_signal(npts, Fs, delay=0.0, seed=42)
    # Normalize for correlation
    import rapidtide.miscmath as tide_math

    sig_normed = tide_math.corrnormalize(signal, detrendorder=3, windowfunc="hamming")
    thedata = np.zeros((2, npts), dtype=float)
    thedata[0, :] = sig_normed
    thedata[1, :] = sig_normed

    maxcorr, maxtime, maskval, failreason = lf.getcorrloc(
        thedata,
        0,
        1,
        Fs,
        dofit=False,
        debug=debug,
    )
    if debug:
        print(f"  maxcorr={maxcorr:.4f}, maxtime={maxtime:.4f}")
    assert maxcorr > 0.9, f"Expected high correlation, got {maxcorr}"
    assert abs(maxtime) < 0.5, f"Expected lag near 0, got {maxtime}"
    assert maskval == 1
    assert failreason == 0


def getcorrloc_delayed_signals(debug=False):
    """Test getcorrloc with delayed signals returns correct lag."""
    if debug:
        print("getcorrloc_delayed_signals")
    Fs = 10.0
    npts = 500
    delay = 1.0  # 1 second delay
    sig1 = _make_broadband_signal(npts, Fs, delay=0.0, seed=42)
    sig2 = _make_broadband_signal(npts, Fs, delay=delay, seed=42)

    import rapidtide.miscmath as tide_math

    thedata = np.zeros((2, npts), dtype=float)
    thedata[0, :] = tide_math.corrnormalize(sig1, detrendorder=3, windowfunc="hamming")
    thedata[1, :] = tide_math.corrnormalize(sig2, detrendorder=3, windowfunc="hamming")

    maxcorr, maxtime, maskval, failreason = lf.getcorrloc(
        thedata,
        0,
        1,
        Fs,
        dofit=False,
        debug=debug,
    )
    if debug:
        print(f"  maxcorr={maxcorr:.4f}, maxtime={maxtime:.4f}")
    assert maxcorr > 0.5, f"Expected decent correlation, got {maxcorr}"
    # Cross-correlation convention: when sig2 is delayed by +delay relative to sig1,
    # the peak appears at -delay (sig1 leads sig2)
    assert (
        abs(abs(maxtime) - delay) < 1.0 / Fs + 0.1
    ), f"Expected lag magnitude near {delay}, got {maxtime}"


def getcorrloc_zero_signal(debug=False):
    """Test getcorrloc with zero signals returns zeros."""
    if debug:
        print("getcorrloc_zero_signal")
    Fs = 10.0
    npts = 100
    thedata = np.zeros((2, npts), dtype=float)
    maxcorr, maxtime, maskval, failreason = lf.getcorrloc(
        thedata,
        0,
        1,
        Fs,
        dofit=False,
    )
    assert maxcorr == 0.0
    assert maxtime == 0.0
    assert maskval == 0
    assert failreason == 0


def getcorrloc_with_fit(debug=False):
    """Test getcorrloc with dofit=True."""
    if debug:
        print("getcorrloc_with_fit")
    Fs = 10.0
    npts = 500
    sig1 = _make_broadband_signal(npts, Fs, delay=0.0, seed=42)
    sig2 = _make_broadband_signal(npts, Fs, delay=0.5, seed=42)

    import rapidtide.miscmath as tide_math

    thedata = np.zeros((2, npts), dtype=float)
    thedata[0, :] = tide_math.corrnormalize(sig1, detrendorder=3, windowfunc="hamming")
    thedata[1, :] = tide_math.corrnormalize(sig2, detrendorder=3, windowfunc="hamming")

    maxcorr, maxtime, maskval, failreason = lf.getcorrloc(
        thedata,
        0,
        1,
        Fs,
        dofit=True,
        debug=debug,
    )
    if debug:
        print(
            f"  maxcorr={maxcorr:.4f}, maxtime={maxtime:.4f}, maskval={maskval}, failreason={failreason}"
        )
    # With fitting, we should get a result (even if fit fails, we get values back)
    # If the fit succeeds, maskval=1
    # If it fails, maskval=0 but we still get numbers


def getcorrloc_one_zero_signal(debug=False):
    """Test getcorrloc when one signal is all zeros."""
    if debug:
        print("getcorrloc_one_zero_signal")
    Fs = 10.0
    npts = 200
    thedata = np.zeros((2, npts), dtype=float)
    thedata[0, :] = _make_broadband_signal(npts, Fs, seed=42)
    # thedata[1, :] is all zeros

    maxcorr, maxtime, maskval, failreason = lf.getcorrloc(
        thedata,
        0,
        1,
        Fs,
        dofit=False,
    )
    assert maxcorr == 0.0
    assert maxtime == 0.0
    assert maskval == 0


def getcorrloc_search_range(debug=False):
    """Test getcorrloc with custom search ranges."""
    if debug:
        print("getcorrloc_search_range")
    Fs = 10.0
    npts = 500
    sig1 = _make_broadband_signal(npts, Fs, delay=0.0, seed=42)
    sig2 = _make_broadband_signal(npts, Fs, delay=0.0, seed=42)

    import rapidtide.miscmath as tide_math

    thedata = np.zeros((2, npts), dtype=float)
    thedata[0, :] = tide_math.corrnormalize(sig1, detrendorder=3, windowfunc="hamming")
    thedata[1, :] = tide_math.corrnormalize(sig2, detrendorder=3, windowfunc="hamming")

    maxcorr, maxtime, maskval, failreason = lf.getcorrloc(
        thedata,
        0,
        1,
        Fs,
        dofit=False,
        negsearch=5.0,
        possearch=5.0,
    )
    assert maskval == 1


# ==================== preprocdata tests ====================


def preprocdata_basic(debug=False):
    """Test preprocdata returns correct shapes."""
    if debug:
        print("preprocdata_basic")
    nx, ny, nz, nt = 3, 3, 2, 40
    Fs = 0.5
    tr = 2.0
    rng = np.random.RandomState(42)
    fmridata = rng.randn(nx, ny, nz, nt).astype(float) + 100.0
    themask = np.ones((nx, ny, nz), dtype=float)

    theprefilter = tide_filt.NoncausalFilter(filtertype="lfo")

    osfmridata, ostimepoints, oversamptr, numvoxels = lf.preprocdata(
        fmridata,
        themask,
        theprefilter,
        2,
        Fs,
        tr,
        detrendorder=1,
        windowfunc="hamming",
        padseconds=0,
        showprogressbar=False,
    )
    numspatiallocs = nx * ny * nz
    assert osfmridata.shape[0] == numspatiallocs
    assert osfmridata.shape[1] == ostimepoints
    assert oversamptr == tr / 2
    assert numvoxels == numspatiallocs  # all voxels in mask


def preprocdata_with_mask(debug=False):
    """Test preprocdata only processes masked voxels."""
    if debug:
        print("preprocdata_with_mask")
    nx, ny, nz, nt = 4, 4, 2, 40
    Fs = 0.5
    tr = 2.0
    rng = np.random.RandomState(42)
    fmridata = rng.randn(nx, ny, nz, nt).astype(float) + 100.0
    themask = np.zeros((nx, ny, nz), dtype=float)
    # Only mask a few voxels
    themask[1, 1, 0] = 1.0
    themask[2, 2, 1] = 1.0
    themask[0, 3, 0] = 1.0

    theprefilter = tide_filt.NoncausalFilter(filtertype="lfo")

    osfmridata, ostimepoints, oversamptr, numvoxels = lf.preprocdata(
        fmridata,
        themask,
        theprefilter,
        2,
        Fs,
        tr,
        detrendorder=1,
        windowfunc="hamming",
        padseconds=0,
        showprogressbar=False,
    )
    assert numvoxels == 3

    # Unmasked voxels should be all zeros
    numspatiallocs = nx * ny * nz
    mask_flat = themask.reshape(numspatiallocs)
    for i in range(numspatiallocs):
        if mask_flat[i] == 0:
            assert np.all(osfmridata[i, :] == 0.0)


def preprocdata_oversample_factor(debug=False):
    """Test preprocdata with different oversample factors."""
    if debug:
        print("preprocdata_oversample_factor")
    nx, ny, nz, nt = 2, 2, 2, 30
    Fs = 0.5
    tr = 2.0
    rng = np.random.RandomState(42)
    fmridata = rng.randn(nx, ny, nz, nt).astype(float) + 100.0
    themask = np.ones((nx, ny, nz), dtype=float)
    theprefilter = tide_filt.NoncausalFilter(filtertype="lfo")

    for factor in [1, 2, 4]:
        osfmridata, ostimepoints, oversamptr, numvoxels = lf.preprocdata(
            fmridata.copy(),
            themask,
            theprefilter,
            factor,
            Fs,
            tr,
            detrendorder=1,
            windowfunc="hamming",
            padseconds=0,
            showprogressbar=False,
        )
        expected_nt = nt * factor - (factor - 1)
        assert (
            ostimepoints == expected_nt
        ), f"factor={factor}: expected {expected_nt} timepoints, got {ostimepoints}"
        assert abs(oversamptr - tr / factor) < 1e-10


def preprocdata_detrend_order(debug=False):
    """Test preprocdata with different detrend orders."""
    if debug:
        print("preprocdata_detrend_order")
    nx, ny, nz, nt = 2, 2, 2, 40
    Fs = 0.5
    tr = 2.0
    rng = np.random.RandomState(42)
    fmridata = rng.randn(nx, ny, nz, nt).astype(float) + 100.0
    themask = np.ones((nx, ny, nz), dtype=float)
    theprefilter = tide_filt.NoncausalFilter(filtertype="lfo")

    for order in [0, 1, 3]:
        osfmridata, ostimepoints, oversamptr, numvoxels = lf.preprocdata(
            fmridata.copy(),
            themask,
            theprefilter,
            2,
            Fs,
            tr,
            detrendorder=order,
            windowfunc="hamming",
            padseconds=0,
            showprogressbar=False,
        )
        assert osfmridata.shape[0] == nx * ny * nz


# ==================== _get_parser tests ====================


def get_parser_returns_parser(debug=False):
    """Test _get_parser returns an ArgumentParser."""
    if debug:
        print("get_parser_returns_parser")
    import argparse

    parser = lf._get_parser()
    assert isinstance(parser, argparse.ArgumentParser)


def get_parser_defaults(debug=False):
    """Test _get_parser default values."""
    if debug:
        print("get_parser_defaults")
    parser = lf._get_parser()
    args = parser.parse_args(["input.nii.gz", "output_root"])
    assert args.inputfilename == "input.nii.gz"
    assert args.outputroot == "output_root"
    assert args.npasses == lf.DEFAULT_NPASSES
    assert args.radius == lf.DEFAULT_RADIUS
    assert args.minlagdiff == lf.DEFAULT_MINLAGDIFF
    assert args.ampthresh == lf.DEFAULT_AMPTHRESH
    assert args.gausssigma == 0.0
    assert args.oversampfactor == -1
    assert not args.dofit
    assert args.detrendorder == lf.DEFAULT_DETREND_ORDER
    assert args.dosphere
    assert args.showprogressbar
    assert not args.debug


def get_parser_npasses(debug=False):
    """Test _get_parser accepts --npasses."""
    if debug:
        print("get_parser_npasses")
    parser = lf._get_parser()
    args = parser.parse_args(["input.nii.gz", "out", "--npasses", "5"])
    assert args.npasses == 5


def get_parser_radius(debug=False):
    """Test _get_parser accepts --radius."""
    if debug:
        print("get_parser_radius")
    parser = lf._get_parser()
    args = parser.parse_args(["input.nii.gz", "out", "--radius", "7.5"])
    assert args.radius == 7.5


def get_parser_ampthresh(debug=False):
    """Test _get_parser accepts --ampthresh."""
    if debug:
        print("get_parser_ampthresh")
    parser = lf._get_parser()
    args = parser.parse_args(["input.nii.gz", "out", "--ampthresh", "0.5"])
    assert args.ampthresh == 0.5


def get_parser_gausssigma(debug=False):
    """Test _get_parser accepts --gausssigma."""
    if debug:
        print("get_parser_gausssigma")
    parser = lf._get_parser()
    args = parser.parse_args(["input.nii.gz", "out", "--gausssigma", "3.0"])
    assert args.gausssigma == 3.0


def get_parser_oversampfac(debug=False):
    """Test _get_parser accepts --oversampfac."""
    if debug:
        print("get_parser_oversampfac")
    parser = lf._get_parser()
    args = parser.parse_args(["input.nii.gz", "out", "--oversampfac", "4"])
    assert args.oversampfactor == 4


def get_parser_dofit(debug=False):
    """Test _get_parser accepts --dofit."""
    if debug:
        print("get_parser_dofit")
    parser = lf._get_parser()
    args = parser.parse_args(["input.nii.gz", "out", "--dofit"])
    assert args.dofit


def get_parser_detrendorder(debug=False):
    """Test _get_parser accepts --detrendorder."""
    if debug:
        print("get_parser_detrendorder")
    parser = lf._get_parser()
    args = parser.parse_args(["input.nii.gz", "out", "--detrendorder", "1"])
    assert args.detrendorder == 1


def get_parser_nosphere(debug=False):
    """Test _get_parser --nosphere disables sphere."""
    if debug:
        print("get_parser_nosphere")
    parser = lf._get_parser()
    args = parser.parse_args(["input.nii.gz", "out", "--nosphere"])
    assert not args.dosphere


def get_parser_noprogressbar(debug=False):
    """Test _get_parser --noprogressbar disables progress bar."""
    if debug:
        print("get_parser_noprogressbar")
    parser = lf._get_parser()
    args = parser.parse_args(["input.nii.gz", "out", "--noprogressbar"])
    assert not args.showprogressbar


def get_parser_debug(debug=False):
    """Test _get_parser accepts --debug."""
    if debug:
        print("get_parser_debug")
    parser = lf._get_parser()
    args = parser.parse_args(["input.nii.gz", "out", "--debug"])
    assert args.debug


def get_parser_minlagdiff(debug=False):
    """Test _get_parser accepts --minlagdiff."""
    if debug:
        print("get_parser_minlagdiff")
    parser = lf._get_parser()
    args = parser.parse_args(["input.nii.gz", "out", "--minlagdiff", "0.5"])
    assert args.minlagdiff == 0.5


def get_parser_windowfunc(debug=False):
    """Test _get_parser accepts --windowfunc."""
    if debug:
        print("get_parser_windowfunc")
    parser = lf._get_parser()
    args = parser.parse_args(["input.nii.gz", "out", "--windowfunc", "hann"])
    assert args.windowfunc == "hann"


def get_parser_filterband(debug=False):
    """Test _get_parser accepts --filterband."""
    if debug:
        print("get_parser_filterband")
    parser = lf._get_parser()
    args = parser.parse_args(["input.nii.gz", "out", "--filterband", "lfo"])
    assert args.filterband == "lfo"


def get_parser_required_args(debug=False):
    """Test _get_parser requires inputfilename and outputroot."""
    if debug:
        print("get_parser_required_args")
    parser = lf._get_parser()
    try:
        parser.parse_args([])
        assert False, "Should have raised SystemExit"
    except SystemExit:
        pass


# ==================== localflow integration test ====================


def localflow_integration(debug=False):
    """Test localflow end-to-end with small synthetic data."""
    if debug:
        print("localflow_integration")
    with tempfile.TemporaryDirectory() as tmpdir:
        shape = (6, 6, 4, 50)
        tr = 2.0
        filepath, data = _make_4d_nifti(tmpdir, shape=shape, tr=tr)
        outputroot = os.path.join(tmpdir, "output")

        # Build args via parser and override
        parser = lf._get_parser()
        args = parser.parse_args([filepath, outputroot])
        # Use small parameters for speed
        args.npasses = 3
        args.radius = 3.0
        args.gausssigma = 0.0
        args.oversampfactor = 1
        args.detrendorder = 1
        args.showprogressbar = False
        args.dofit = False
        args.debug = debug
        args.ampthresh = 0.1

        # postprocess filter options (needed by localflow)
        _, theprefilter = pf.postprocessfilteropts(args)

        lf.localflow(args)

        # Check output files exist
        assert os.path.exists(outputroot + "_mask.nii.gz"), "mask not created"
        assert os.path.exists(outputroot + "_preprocdata.nii.gz"), "preprocdata not created"
        assert os.path.exists(outputroot + "_corrcoeffs.nii.gz"), "corrcoeffs not created"
        assert os.path.exists(outputroot + "_delays.nii.gz"), "delays not created"
        assert os.path.exists(outputroot + "_corrvalid.nii.gz"), "corrvalid not created"
        assert os.path.exists(outputroot + "_failreason.nii.gz"), "failreason not created"
        assert os.path.exists(outputroot + "_targetdelay.nii.gz"), "targetdelay not created"
        assert os.path.exists(outputroot + "_numneighbors.nii.gz"), "numneighbors not created"
        assert os.path.exists(outputroot + "_maxtime.nii.gz"), "maxtime not created"
        assert os.path.exists(outputroot + "_formattedruntimings.txt"), "timings not created"
        assert os.path.exists(outputroot + "_neighbors"), "neighbors not created"
        assert os.path.exists(outputroot + "_indexlist"), "indexlist not created"

        # Read back and verify shapes
        _, mask_data, _, maskdims, _ = tide_io_readfromnifti(outputroot + "_mask.nii.gz")
        assert mask_data.shape == shape[:3]

        _, targetdelay_data, _, _, _ = tide_io_readfromnifti(outputroot + "_targetdelay.nii.gz")
        assert targetdelay_data.shape == (shape[0], shape[1], shape[2], args.npasses)

        _, maxtime_data, _, _, _ = tide_io_readfromnifti(outputroot + "_maxtime.nii.gz")
        assert maxtime_data.shape == shape[:3]

        if debug:
            print("  Integration test passed - all output files verified")


def localflow_with_gausssigma(debug=False):
    """Test localflow with Gaussian spatial smoothing."""
    if debug:
        print("localflow_with_gausssigma")
    with tempfile.TemporaryDirectory() as tmpdir:
        shape = (6, 6, 4, 40)
        tr = 2.0
        filepath, data = _make_4d_nifti(tmpdir, shape=shape, tr=tr)
        outputroot = os.path.join(tmpdir, "output")

        parser = lf._get_parser()
        args = parser.parse_args([filepath, outputroot])
        args.npasses = 2
        args.radius = 3.0
        args.gausssigma = 2.0
        args.oversampfactor = 1
        args.detrendorder = 1
        args.showprogressbar = False
        args.dofit = False
        args.debug = debug
        args.ampthresh = 0.1

        _, theprefilter = pf.postprocessfilteropts(args)

        lf.localflow(args)

        assert os.path.exists(outputroot + "_mask.nii.gz")
        assert os.path.exists(outputroot + "_maxtime.nii.gz")
        if debug:
            print("  Gausssigma test passed")


# We need tide_io for reading back NIfTI files in integration tests
def tide_io_readfromnifti(filepath):
    """Helper to read NIfTI files for verification."""
    import rapidtide.io as tide_io

    return tide_io.readfromnifti(filepath)


# ==================== Main test function ====================


def test_localflow(debug=False):
    # xyz2index tests
    if debug:
        print("Running xyz2index tests")
    xyz2index_basic(debug=debug)
    xyz2index_corners(debug=debug)
    xyz2index_out_of_bounds(debug=debug)
    xyz2index_boundary(debug=debug)
    xyz2index_nonsquare(debug=debug)

    # index2xyz tests
    if debug:
        print("Running index2xyz tests")
    index2xyz_basic(debug=debug)
    index2xyz_nonsquare(debug=debug)
    index2xyz_roundtrip(debug=debug)
    index2xyz_sequential(debug=debug)

    # getcorrloc tests
    if debug:
        print("Running getcorrloc tests")
    getcorrloc_identical_signals(debug=debug)
    getcorrloc_delayed_signals(debug=debug)
    getcorrloc_zero_signal(debug=debug)
    getcorrloc_with_fit(debug=debug)
    getcorrloc_one_zero_signal(debug=debug)
    getcorrloc_search_range(debug=debug)

    # preprocdata tests
    if debug:
        print("Running preprocdata tests")
    preprocdata_basic(debug=debug)
    preprocdata_with_mask(debug=debug)
    preprocdata_oversample_factor(debug=debug)
    preprocdata_detrend_order(debug=debug)

    # _get_parser tests
    if debug:
        print("Running _get_parser tests")
    get_parser_returns_parser(debug=debug)
    get_parser_defaults(debug=debug)
    get_parser_npasses(debug=debug)
    get_parser_radius(debug=debug)
    get_parser_ampthresh(debug=debug)
    get_parser_gausssigma(debug=debug)
    get_parser_oversampfac(debug=debug)
    get_parser_dofit(debug=debug)
    get_parser_detrendorder(debug=debug)
    get_parser_nosphere(debug=debug)
    get_parser_noprogressbar(debug=debug)
    get_parser_debug(debug=debug)
    get_parser_minlagdiff(debug=debug)
    get_parser_windowfunc(debug=debug)
    get_parser_filterband(debug=debug)
    get_parser_required_args(debug=debug)

    # localflow integration tests
    if debug:
        print("Running localflow integration tests")
    localflow_integration(debug=debug)
    localflow_with_gausssigma(debug=debug)


if __name__ == "__main__":
    test_localflow(debug=True)
